from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada.fem.shapes import definitions as shape_def


class AbaqusDefaultShellTypes:
    def __init__(self):
        self.TRIANGLE = "S3"
        self.TRIANGLE6 = "STRI65"
        self.TRIANGLE7 = "S7"
        self.QUAD = "S4R"
        self.QUAD8 = "S8R"


@dataclass
class AbaqusDefaultSolidTypes:
    HEXAHEDRON = "C3D8"
    HEXAHEDRON20 = "C3D20R"
    HEXAHEDRON27 = "C3D27"
    TETRA = "C3D4"
    TETRA10 = "C3D10"
    PYRAMID5 = "C3D5"
    PRISM6 = "C3D6"
    PRISM15 = "C3D15"


@dataclass
class AbaqusDefaultLineTypes:
    LINE = "B31"
    LINE3 = "B32"


class AbaqusDefaultElemTypes:
    def __init__(self):
        self.LINE = AbaqusDefaultLineTypes()
        self.SHELL = AbaqusDefaultShellTypes()
        self.SOLID = AbaqusDefaultSolidTypes()

    def get_element_type(self, el_type: shape_def.LineShapes | shape_def.ShellShapes | shape_def.SolidShapes) -> str:
        from ada.fem.shapes import ElemType
        from ada.fem.shapes.definitions import ConnectorTypes, MassTypes, ShapeResolver

        if isinstance(el_type, (MassTypes, ConnectorTypes)):
            return str(el_type.value)

        type_group = ShapeResolver.to_geom_repr(el_type)

        type_map = {
            ElemType.LINE: self.LINE,
            ElemType.SHELL: self.SHELL,
            ElemType.SOLID: self.SOLID,
        }

        res = getattr(type_map[type_group], el_type.value.upper(), None)

        if res is None:
            raise ValueError(f'Unrecognized element type "{el_type}"')

        return res


class AbaqusPointTypes:
    spring1n = ["SPRING1"]
    masses = ["MASS", "ROTARYI"]

    all = [spring1n, masses]



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.core.constants import O, X, Z
from ada.core.utils import Counter
from ada.ifc.utils import create_guid, create_local_placement, write_elem_property_sets

from .write_shapes import generate_parametric_solid

if TYPE_CHECKING:
    from ada import Penetration, Shape

pen_counter = Counter(prefix="P")


def generate_ifc_opening(primitive: Penetration | Shape):
    from ada import Penetration, Shape

    if primitive.parent is None:
        raise ValueError("This penetration has no parent")
    pen_name = f"{primitive.name}_{next(pen_counter)}"
    a = primitive.get_assembly()
    parent_part = primitive.parent.parent
    f = a.ifc_store.f

    geom_parent = f.by_guid(parent_part.guid)
    owner_history = a.ifc_store.owner_history

    # Create and associate an opening for the window in the wall
    opening_placement = create_local_placement(f, O, Z, X, relative_to=geom_parent.ObjectPlacement)
    if isinstance(primitive, Penetration):
        prim = primitive.primitive
    elif issubclass(type(primitive), Shape):
        prim = primitive
    else:
        raise NotImplementedError()
    opening_shape = generate_parametric_solid(prim, f)

    opening_element = f.create_entity(
        "IfcOpeningElement",
        GlobalId=create_guid(),
        OwnerHistory=owner_history,
        Name=pen_name,
        Description=pen_name + " (Opening)",
        ObjectPlacement=opening_placement,
        Representation=opening_shape,
    )

    write_elem_property_sets(primitive.metadata, opening_element, f, owner_history)

    return opening_element



import pathlib
import xml.etree.ElementTree as ET

from ada import Part
from ada.config import get_logger
from ada.fem.formats.sesam.xml.read.helpers import (
    apply_mass_density_factors,
    yield_plate_elems_to_plate,
)
from ada.fem.formats.sesam.xml.read.read_bcs import get_boundary_conditions
from ada.fem.formats.sesam.xml.read.read_beams import el_to_beam
from ada.fem.formats.sesam.xml.read.read_joints import get_joints
from ada.fem.formats.sesam.xml.read.read_masses import get_masses
from ada.fem.formats.sesam.xml.read.read_materials import get_materials
from ada.fem.formats.sesam.xml.read.read_sections import get_sections
from ada.fem.formats.sesam.xml.read.read_sets import get_sets
from ada.fem.formats.sesam.xml.sat_helpers import write_xml_sat_text_to_file
from ada.sat.factory import SatReaderFactory

logger = get_logger()


class GxmlStore:
    def __init__(self, xml_path: pathlib.Path):
        if isinstance(xml_path, str):
            xml_path = pathlib.Path(xml_path).resolve().absolute()

        self.sat_file = xml_path.with_suffix(".sat")

        if not self.sat_file.exists():
            logger.info("SAT file does not exist. Creating SAT file")
            write_xml_sat_text_to_file(xml_file=xml_path, out_file=self.sat_file)
        elif self.sat_file.exists() and self.sat_file.lstat().st_ctime < xml_path.lstat().st_ctime:
            logger.info("XML file is newer than SAT file. Updating SAT file")
            write_xml_sat_text_to_file(xml_file=xml_path, out_file=self.sat_file)

        self.xml_root = ET.parse(str(xml_path)).getroot()
        self.sat_factory = SatReaderFactory(self.sat_file)

        model = self.xml_root.find(".//model")
        p = Part(model.attrib["name"])
        self.p = p
        p._sections = get_sections(self.xml_root, p)
        p._materials = get_materials(self.xml_root, p)

    def iter_geometry_from_xml(self):
        yield from self.iter_beams_from_xml()
        yield from self.iter_plates_from_xml()

    def iter_beams_from_xml(self):
        p = self.p

        for bm in self.xml_root.iterfind(".//straight_beam"):
            yield from el_to_beam(bm, p)

        for curved_bm in self.xml_root.iterfind(".//curved_beam"):
            yield from el_to_beam(curved_bm, p)

    def iter_plates_from_xml(self):
        sat_d = {name: points for name, points in self.sat_factory.iter_flat_plates()}
        thick_map = dict()
        for thickn in self.xml_root.iterfind(".//thickness"):
            res = thickn.find(".//constant_thickness")
            thick_map[thickn.attrib["name"]] = float(res.attrib["th"])

        for fp in self.xml_root.iterfind(".//flat_plate"):
            yield from yield_plate_elems_to_plate(fp, self.p, sat_d, thick_map)

        for fp in self.xml_root.iterfind(".//curved_shell"):
            yield from yield_plate_elems_to_plate(fp, self.p, sat_d, thick_map)

    def to_part(self, extract_joints=False) -> Part:
        from ada.concepts.containers import Beams, Plates

        p = self.p
        p._plates = Plates(self.iter_plates_from_xml(), parent=p)
        p._beams = Beams(self.iter_beams_from_xml(), parent=p)
        for bm in p.beams:
            p.nodes.add(bm.n1)
            p.nodes.add(bm.n2)
        p._groups = get_sets(self.xml_root, p)
        if extract_joints is True:
            p._connections = get_joints(self.xml_root, p)

        p.fem.bcs += get_boundary_conditions(self.xml_root, p)
        p.fem.masses.update(get_masses(self.xml_root, p))

        all_plates = len(p.plates)
        all_beams = len(p.beams)
        all_joints = len(p.connections)

        apply_mass_density_factors(self.xml_root, p)

        print(f"Finished importing Genie XML (beams={all_beams}, plates={all_plates}, joints={all_joints})")
        return p



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Union

_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

if TYPE_CHECKING:
    from ada import FEM, Part
    from ada.fem import FemSet, Surface


class AbaFF:
    """Abaqus Fortran Flags. A class designed to aid in building regex searched for Abaqus flags."""

    def __init__(self, flag, args, subflags=None, nameprop=None):
        """
        :param flag: Main flag. *<FLAG>
        :param args: arguments. Tuple of arguments
        :param subflags:
        :param nameprop:


        type of arguments:

        >: up to

        """
        self._flag = flag
        self._subflags = subflags if subflags is not None else []
        self._nameprop = nameprop
        self._args = args

    def _create_regex(self, flag, nameprop, args):
        regstr = ""
        if nameprop is not None:
            regstr += rf"\*\*\s*{nameprop[0]}:\s*(?P<{nameprop[1]}>.*?)\n"
        regstr += rf"\*{flag}"
        for i, arg in enumerate(args):
            for j, fl in enumerate(arg):
                subfl = fl.replace("=", "").replace("|", "").replace(">", "")
                exact = True if "==" in fl else False
                if exact:
                    equality = False
                else:
                    equality = True if "=" in fl else False
                optional = True if "|" in fl else False
                uptostar = True if ">" in fl else False

                last_char = regstr[-10:]
                regstr += "(?:"

                if r"(?:\n|)\s*" != last_char:
                    regstr += ","

                clean_name = subfl.replace(" ", "_").replace(r"\*", "")

                if equality:
                    regstr += rf"\s*{subfl}=(?P<{clean_name}>"
                else:
                    regstr += rf"\s*(?P<{clean_name}>"

                if uptostar:
                    regstr += r"(?:(?!\*).)*"
                else:
                    if exact is True:
                        regstr += subfl
                    else:
                        regstr += ".*?"
                regstr += ")"

                if j + 1 == len(arg):
                    regstr += "$"

                if optional:
                    regstr += "|"
                    if j + 1 == len(arg):
                        regstr += "$"

                regstr += ")"
            regstr += r"(?:\n|)\s*"
        return regstr

    @property
    def regstr(self):
        regstr = self._create_regex(self._flag, self._nameprop, self._args)
        for v in self._subflags:
            clean_name = v[0].replace(" ", "_").replace(r"\*", "")
            regstr += f"(:?(?P<{clean_name}>)"
            regstr += self._create_regex(v[0], None, v[1])
            regstr += ")"
        return regstr

    @property
    def regex(self):
        return re.compile(self.regstr, re.IGNORECASE | re.MULTILINE | re.DOTALL)


def list_cleanup(membulkstr):
    return membulkstr.replace(",\n", ",").replace("\n", ",")


def is_set_in_part(part: Part, set_name: str, set_type) -> Union[FemSet, Surface]:
    set_map = {"nset": part.fem.nsets, "elset": part.fem.elsets, "surface": part.fem.surfaces}
    id_map = {"nset": part.fem.nodes, "elset": part.fem.elements}

    if str.isnumeric(set_name):
        _id = int(set_name)
        return id_map[set_type].from_id(_id)

    if set_name in set_map[set_type].keys():
        return set_map[set_type][set_name]

    raise ValueError()


def get_set_from_assembly(set_str: str, fem: "FEM", set_type) -> Union["FemSet", "Surface"]:
    res = set_str.split(".")

    if len(res) == 1:
        local_set_map = {"nset": fem.nsets, "elset": fem.elsets, "surface": fem.surfaces}
        set_name = res[0]
        return local_set_map[set_type][set_name]

    set_name = res[1]
    p_name = res[0]

    if str.isnumeric(set_name):
        num_id = int(set_name)
        local_id_map = {"nset": fem.nodes.from_id, "elset": fem.elements.from_id}
        if p_name == fem.name:
            return local_id_map[set_type](num_id)
        for part in fem.parent.get_all_parts_in_assembly():
            if p_name == part.fem.instance_name:
                r = is_set_in_part(part, set_name, set_type)
                if r is not None:
                    return r
    else:
        local_set_map = {"nset": fem.nsets, "elset": fem.elsets, "surface": fem.surfaces}

        if p_name == fem.name:
            return local_set_map[p_name]
        for part in fem.parent.get_all_parts_in_assembly():
            if p_name == part.fem.instance_name:
                r = is_set_in_part(part, set_name, set_type)
                if r is not None:
                    return r
    raise ValueError(f'No {set_type} "{set_str}" was found')



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from ada.config import Settings as _Settings
from ada.config import get_logger
from ada.ifc.utils import create_guid

from .changes import ChangeAction
from .units import Units

if TYPE_CHECKING:
    from ada import Assembly, Part
    from ada.ifc.store import IfcStore

logger = get_logger()


class Root:
    UNITS: Units = Units

    def __init__(
        self,
        name,
        guid=None,
        metadata=None,
        units: Units | str = Units.M,
        parent=None,
        ifc_store: IfcStore = None,
        change_type: ChangeAction = ChangeAction.ADDED,
    ):
        self.name = name
        self.parent = parent
        self.change_type = change_type
        self.guid = create_guid() if guid is None else guid

        if isinstance(units, str):
            units = Units.from_str(units)
        self._units = units
        self._metadata = metadata if metadata is not None else dict()
        self._ifc_store = ifc_store

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if _Settings.convert_bad_names:
            logger.debug("Converting bad name")
            value = value.replace("/", "_").replace("=", "")
            if str.isnumeric(value[0]):
                value = "ADA_" + value

        if "/" in value:
            logger.debug(f'Character "/" found in {value}')

        self._name = value.strip()

    @property
    def guid(self):
        return self._guid

    @guid.setter
    def guid(self, value):
        if value is None:
            raise ValueError("guid cannot be None")
        if hasattr(self, "_guid") and self.guid is not None:
            existing_guid = self.guid
            a = self.get_assembly()
            original_el = a.ifc_store.f.by_guid(existing_guid)
            original_el.GlobalId = value
        self._guid = value

    @property
    def parent(self) -> Part:
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        raise NotImplementedError("Assigning units is not yet represented for this object")

    def get_assembly(self) -> Union[Assembly, Part]:
        from ada import Assembly

        for ancestor in self.get_ancestors():
            if isinstance(ancestor, Assembly):
                return ancestor
        logger.info("No Assembly found in ancestry. Returning self")
        return self

    def get_ancestors(self) -> List[Union[Part, Assembly]]:
        ancestry = [self]
        current = self
        while current.parent is not None:
            ancestry.append(current.parent)
            current = current.parent
        return ancestry

    def remove(self):
        """Remove this element/part from assembly/part"""
        from ada import Beam, Part, Plate, Section, Shape

        if self.parent is None:
            logger.error(f"Unable to delete {self.name} as it does not have a parent")
            return

        if issubclass(type(self), Part):
            self.parent.parts.pop(self.name)
        elif issubclass(type(self), Shape):
            self.parent.shapes.pop(self.parent.shapes.index(self))
        elif isinstance(self, Beam):
            self.parent.beams.remove(self)
        elif isinstance(self, Plate):
            self.parent.plates.remove(self)
        elif isinstance(self, Section):
            logger.warning("Section removal is not yet supported")
        else:
            raise NotImplementedError()



from __future__ import annotations

from dataclasses import dataclass

import ifcopenshell
from ifcopenshell.util.element import get_psets

from ada import Assembly, Part
from ada.base.ifc_types import SpatialTypes
from ada.config import get_logger
from ada.ifc.store import IfcStore

from .reader_utils import get_ifc_property_sets, get_parent, resolve_name

logger = get_logger()


def valid_spatial_classes(product: ifcopenshell.entity_instance):
    is_ok_class = SpatialTypes.is_valid_spatial_type(product.is_a())
    has_no_geom = product.Representation is None

    if is_ok_class is True and has_no_geom is True:
        return True

    if is_ok_class is True:
        logger.info(f"{product=}-> {has_no_geom=}")
        return True

    return False


@dataclass
class PartImporter:
    ifc_store: IfcStore

    def get_parent(self, product: ifcopenshell.entity_instance) -> Part | Assembly | None:
        pp = get_parent(product)
        pp_name = pp.Name
        if pp_name is None:
            pp_name = resolve_name(get_psets(pp), pp)
        if pp_name is None:
            return None
        return self.ifc_store.assembly.get_by_name(pp_name)

    def load_hierarchies(self) -> None:
        for product in filter(valid_spatial_classes, self.ifc_store.f.by_type("IfcProduct")):
            class_type = SpatialTypes.from_str(product.is_a())
            if class_type == SpatialTypes.IfcSite:
                self.update_assembly(product)
                continue
            new_part = self.import_ifc_hierarchy(product)
            parent = self.get_parent(product)
            if parent is None:
                if new_part.name not in self.ifc_store.assembly.parts.keys():
                    self.ifc_store.assembly.add_part(new_part)
            elif isinstance(parent, Part) is False:
                raise NotImplementedError()
            else:
                parent.add_part(new_part)

    def import_ifc_hierarchy(self, product: ifcopenshell.entity_instance) -> Part:
        props = get_ifc_property_sets(product)
        name = product.Name
        if name is None:
            logger.debug(f'Name was not found for the IFC element "{product}". Will look for ref to name in props')
            name = resolve_name(props, product)

        ifc_class = SpatialTypes.from_str(product.is_a())

        return Part(
            name,
            metadata=props,
            guid=product.GlobalId,
            ifc_store=self.ifc_store,
            units=self.ifc_store.assembly.units,
            ifc_class=ifc_class,
        )

    def update_assembly(self, product: ifcopenshell.entity_instance):
        props = get_ifc_property_sets(product)
        name = product.Name
        if name is None:
            logger.debug(f'Name was not found for the IFC element "{product}". Will look for ref to name in props')
            name = resolve_name(props, product)

        self.ifc_store.assembly.name = name
        self.ifc_store.assembly._guid = product.GlobalId
        self.ifc_store.assembly.metadata.update(props)



from typing import Iterator

from ada import Beam
from ada.core.utils import Counter


def beams_along_polyline(polyline: list[tuple], bm: Beam, name_gen: Iterator = None) -> list[Beam]:
    beams = []
    ngen = name_gen if name_gen is not None else Counter(prefix="bm")
    for p1, p2 in zip(polyline[:-1], polyline[1:]):
        beams.append(bm.copy_to(p1, p2, next(ngen)))
    return beams



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from OCC.Extend.DataExchange import read_step_file

import ada
from ada.base.types import GeomRepr
from ada.config import get_logger

if TYPE_CHECKING:
    from OCC.Core.TopoDS import TopoDS_Shape

    from ada.base.physical_objects import BackendGeom
    from ada.occ.step.writer import StepWriter

logger = get_logger()


class OCCStore:
    def __init__(self, step_file: str | pathlib.Path = None):
        self.step_file = step_file

    def iter_shapes(self):
        for shp in read_step_file(self.step_file, as_compound=False):
            yield shp

    @staticmethod
    def get_writer() -> StepWriter:
        from ada.occ.step.writer import StepWriter

        return StepWriter("AdaStep")

    @staticmethod
    def shape_iterator(
        part: ada.Part | BackendGeom, geom_repr: GeomRepr = GeomRepr.SOLID
    ) -> tuple[BackendGeom, TopoDS_Shape]:
        if isinstance(geom_repr, str):
            geom_repr = GeomRepr.from_str(geom_repr)

        def safe_geom(obj_):
            try:
                if geom_repr == GeomRepr.SOLID:
                    return obj_.solid()
                elif geom_repr == GeomRepr.SHELL:
                    return obj_.shell()
            except RuntimeError as e:
                logger.warning(f"Failed to add shape {obj.name} due to {e}")
                return None

        if isinstance(part, (ada.Part, ada.Assembly)):
            for obj in part.get_all_physical_objects(pipe_to_segments=True):
                if isinstance(geom_repr, str):
                    geom_repr = GeomRepr.from_str(geom_repr)

                if issubclass(type(obj), ada.Shape):
                    geom = safe_geom(obj)
                elif isinstance(obj, (ada.Beam, ada.Plate, ada.Wall)):
                    geom = safe_geom(obj)
                elif isinstance(obj, (ada.PipeSegStraight, ada.PipeSegElbow)):
                    geom = safe_geom(obj)
                else:
                    logger.error(f"Geometry type {type(obj)} not yet implemented")
                    geom = None

                if geom is None:
                    continue

                yield obj, geom

        else:
            yield part, safe_geom(part)



from datetime import datetime

import ifcopenshell

from ada.concepts.user import User
from ada.ifc.read.reader_utils import get_org, get_person


def create_owner_history_from_user(user: User, f: ifcopenshell.file) -> ifcopenshell.entity_instance:
    actor = None
    for ar in f.by_type("IfcActorRole"):
        if ar.Role == user.role.upper():
            actor = ar
            break

    if actor is None:
        actor = f.create_entity("IfcActorRole", Role=user.role.upper(), UserDefinedRole=None, Description=None)

    user_props = dict(
        Identification=user.user_id,
        FamilyName=user.family_name,
        GivenName=user.given_name,
        MiddleNames=user.middle_names,
        PrefixTitles=user.prefix_titles,
        SuffixTitles=user.suffix_titles,
    )

    person = get_person(f, user.user_id)
    if person is None:
        person = f.create_entity("IfcPerson", **user_props, Roles=(actor,))

    organization = get_org(f, user.org_id)
    if organization is None:
        organization = f.create_entity(
            "IfcOrganization",
            Identification=user.org_id,
            Name=user.org_name,
            Description=user.org_description,
        )

    p_o = None
    for po in f.by_type("IfcPersonAndOrganization"):
        if po.TheOrganization != organization:
            continue
        p_o = po
        break

    if p_o is None:
        p_o = f.create_entity("IfcPersonAndOrganization", person, organization)

    app_name = "ADA"
    application = None
    for app in f.by_type("IfcApplication"):
        if app.ApplicationFullName != app_name:
            continue
        application = app
        break

    if application is None:
        application = f.create_entity("IfcApplication", organization, "XXX", "ADA", "ADA")

    timestamp = int(datetime.now().timestamp())

    owner_history = None
    for oh in f.by_type("IfcOwnerHistory"):
        if oh.OwningUser != p_o:
            continue
        if oh.OwningApplication != application:
            continue
        oh.LastModifiedDate = timestamp
        owner_history = oh
        break

    if owner_history is None:
        owner_history = f.create_entity(
            "IfcOwnerHistory",
            OwningUser=p_o,
            OwningApplication=application,
            State="READWRITE",
            ChangeAction=None,
            LastModifiedDate=None,
            LastModifyingUser=p_o,
            LastModifyingApplication=application,
            CreationDate=timestamp,
        )

    return owner_history



from __future__ import annotations

from typing import Union

import numpy as np

from ada import Part
from ada.base.physical_objects import BackendGeom
from ada.base.units import Units
from ada.concepts.curves import CurvePoly
from ada.concepts.primitives import PrimBox
from ada.concepts.transforms import Placement
from ada.core.vector_utils import unit_vector


class WallJustification:
    CENTER = "CENTER"
    LEFT = "LEFT"
    RIGHT = "RIGHT"

    all = [CENTER, LEFT, RIGHT]


class Wall(BackendGeom):
    TYPES_JUSL = WallJustification

    """
    A wall object representing

    :param points: Points making up wall
    :param height: Height
    :param thickness: Thickness
    :param origin: Origin
    :param offset: Wall offset from points making up the wall centerline. Accepts float | CENTER | LEFT | RIGHT
    """

    def __init__(
        self,
        name,
        points,
        height,
        thickness,
        placement=Placement(),
        offset=TYPES_JUSL.CENTER,
        metadata=None,
        colour=None,
        units=Units.M,
        guid=None,
        opacity=1.0,
    ):
        super().__init__(name, guid=guid, metadata=metadata, units=units, colour=colour, opacity=opacity)

        self._name = name
        self.placement = placement
        self.colour = colour
        new_points = []
        for p in points:
            np_ = [float(c) for c in p]
            if len(np_) == 2:
                np_ += [0.0]
            new_points.append(tuple(np_))
        self._points = new_points
        self._segments = list(zip(self._points[:-1], self.points[1:]))
        self._height = height
        self._thickness = thickness
        self._openings = []
        self._doors = []
        self._inserts = []
        if type(offset) is str:
            if offset not in Wall.TYPES_JUSL.all:
                raise ValueError(f'Unknown string input "{offset}" for offset')
            if offset == Wall.TYPES_JUSL.CENTER:
                self._offset = 0.0
            elif offset == Wall.TYPES_JUSL.LEFT:
                self._offset = -self._thickness / 2
            else:  # offset = RIGHT
                self._offset = self._thickness / 2
        else:
            if type(offset) not in (float, int):
                raise ValueError("Offset can only be string or float, int")
            self._offset = offset

    def add_insert(self, insert: "WallInsert", wall_segment: int, off_x, off_z):
        """

        :param insert:
        :param wall_segment:
        :param off_x:
        :param off_z:
        :return:
        """
        from OCC.Extend.ShapeFactory import get_oriented_boundingbox

        xvec, yvec, zvec = self.get_segment_props(wall_segment)
        p1, p2 = self._segments[wall_segment]

        start = p1 + yvec * (self._thickness / 2 + self.offset) + xvec * off_x + zvec * off_z
        insert._depth = self._thickness
        insert.placement = Placement(origin=start, xdir=xvec, ydir=zvec, zdir=yvec)

        frame = insert.shapes[0]
        center, dim, oobb_shp = get_oriented_boundingbox(frame.geom())
        x, y, z = center.X(), center.Y(), center.Z()
        dx, dy, dz = dim[0], dim[1], dim[2]

        x0 = x - abs(dx / 2)
        y0 = y - abs(dy / 2)
        z0 = z - abs(dz / 2)

        x1 = x + abs(dx / 2)
        y1 = y + abs(dy / 2)
        z1 = z + abs(dz / 2)

        self._inserts.append(insert)
        self._openings.append([wall_segment, insert, (x0, y0, z0), (x1, y1, z1)])

        tol = 0.4
        wi = insert

        p1 = wi.placement.origin - yvec * (wi.depth / 2 + tol)
        p2 = wi.placement.origin + yvec * (wi.depth / 2 + tol) + xvec * wi.width + zvec * wi.height

        self._penetrations.append(PrimBox("my_pen", p1, p2))

    def get_segment_props(self, wall_segment):
        """

        :param wall_segment:
        :return:
        """
        if wall_segment > len(self._segments):
            raise ValueError(f"Wall segment id should be equal or less than {len(self._segments)}")

        p1, p2 = self._segments[wall_segment]
        xvec = unit_vector(np.array(p2) - np.array(p1))
        zvec = np.array([0, 0, 1])
        yvec = unit_vector(np.cross(zvec, xvec))

        return xvec, yvec, zvec

    def _generate_ifc_elem(self):
        from ada.ifc.write.write_wall import write_ifc_wall

        return write_ifc_wall(self)

    @property
    def inserts(self):
        return self._inserts

    @property
    def height(self):
        return self._height

    @property
    def thickness(self):
        return self._thickness

    @property
    def placement(self) -> Placement:
        return self._placement

    @placement.setter
    def placement(self, value: Placement):
        self._placement = value

    @property
    def points(self):
        return self._points

    @property
    def offset(self) -> Union[float, str]:
        return self._offset

    @property
    def extrusion_area(self):
        from ada.core.vector_utils import calc_yvec, intersect_calc, is_parallel

        area_points = []
        vpo = [np.array(p) for p in self.points]
        p2 = None
        yvec = None
        prev_xvec = None
        prev_yvec = None
        zvec = np.array([0, 0, 1])

        # Inner line
        for p1, p2 in zip(vpo[:-1], vpo[1:]):
            xvec = p2 - p1
            yvec = unit_vector(calc_yvec(xvec, zvec))
            new_point = p1 + yvec * (self._thickness / 2) + yvec * self.offset
            if prev_xvec is not None:
                if is_parallel(xvec, prev_xvec) is False:
                    prev_p = area_points[-1]
                    # next_point = p2 + yvec * (self._thickness / 2) + yvec * self.offset
                    # c_p = prev_yvec * (self._thickness / 2) + prev_yvec * self.offset
                    AB = prev_xvec
                    CD = xvec
                    s, t = intersect_calc(prev_p, new_point, AB, CD)
                    sAB = prev_p + s * AB
                    new_point = sAB
            area_points.append(new_point)
            prev_xvec = xvec
            prev_yvec = yvec

        # Add last point
        area_points.append((p2 + yvec * (self._thickness / 2) + yvec * self.offset))
        area_points.append((p2 - yvec * (self._thickness / 2) + yvec * self.offset))

        reverse_points = []
        # Outer line
        prev_xvec = None
        prev_yvec = None
        for p1, p2 in zip(vpo[:-1], vpo[1:]):
            xvec = p2 - p1
            yvec = unit_vector(calc_yvec(xvec, zvec))
            new_point = p1 - yvec * (self._thickness / 2) + yvec * self.offset
            if prev_xvec is not None:
                if is_parallel(xvec, prev_xvec) is False:
                    prev_p = reverse_points[-1]
                    c_p = prev_yvec * (self._thickness / 2) - prev_yvec * self.offset
                    new_point -= c_p
            reverse_points.append(new_point)
            prev_xvec = xvec
            prev_yvec = yvec

        reverse_points.reverse()
        area_points += reverse_points

        new_points = []
        for p in area_points:
            new_points.append(tuple([float(c) for c in p]))

        return new_points

    @property
    def openings_extrusions(self):
        from ada.concepts.spatial import Part

        op_extrudes = []
        if self.units == Units.M:
            tol = 0.4
        else:
            tol = 400
        for op in self._openings:
            ws, wi, mi, ma = op
            xvec, yvec, zvec = self.get_segment_props(ws)
            assert issubclass(type(wi), Part)
            p1 = wi.placement.origin - yvec * (wi.depth / 2 + tol)
            p2 = p1 + yvec * (wi.depth + tol * 2)
            p3 = p2 + xvec * wi.width
            p4 = p3 - yvec * (wi.depth + tol * 2)
            op_extrudes.append([p1.tolist(), p2.tolist(), p3.tolist(), p4.tolist(), p1.tolist()])
        return op_extrudes

    @property
    def metadata(self):
        return self._metadata

    @property
    def shell(self):
        poly = CurvePoly(points3d=self.extrusion_area, parent=self)
        return poly.face

    @property
    def solid(self):
        from ada.occ.utils import apply_penetrations

        poly = CurvePoly(points3d=self.extrusion_area, parent=self)

        geom = apply_penetrations(poly.make_extruded_solid(self.height), self.penetrations)

        return geom

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            scale_factor = Units.get_scale_factor(self._units, value)
            self._height *= scale_factor
            self._thickness *= scale_factor
            self._offset *= scale_factor
            self.placement.origin = np.array([x * scale_factor for x in self.placement.origin])
            self._points = [tuple([x * scale_factor for x in p]) for p in self.points]
            self._segments = list(zip(self._points[:-1], self.points[1:]))
            for pen in self._penetrations:
                pen.units = value
            for opening in self._openings:
                opening[2] = tuple([x * scale_factor for x in opening[2]])
                opening[3] = tuple([x * scale_factor for x in opening[3]])

            for insert in self._inserts:
                insert.units = value

            self._units = value

    def __repr__(self):
        return f"Wall({self.name})"


class WallInsert(Part):
    def __init__(self, name, width, height, depth, **kwargs):
        super(WallInsert, self).__init__(name, **kwargs)
        self._width = width
        self._height = height
        self._depth = depth
        self._is_built = False

    def build_geom(self):
        raise NotImplementedError()

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def depth(self):
        return self._depth

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, value):
        self._placement = value
        self.build_geom()
        self._is_built = True



from __future__ import annotations

import mmap
import os
import pathlib
import re
from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, Dict, List, Union

import numpy as np

from ada.concepts.containers import Nodes
from ada.concepts.points import Node
from ada.concepts.transforms import Rotation, Transform
from ada.config import get_logger
from ada.core.utils import Counter
from ada.fem import (
    Bc,
    Constraint,
    FemSet,
    Interaction,
    InteractionProperty,
    PredefinedField,
    Surface,
)
from ada.fem.containers import FemSets
from ada.fem.formats.utils import str_to_int
from ada.fem.interactions import ContactTypes
from ada.fem.shapes import ElemType

from . import cards
from .helper_utils import _re_in, get_set_from_assembly, list_cleanup
from .read_elements import get_elem_from_bulk_str, update_connector_data
from .read_masses import get_mass_from_bulk
from .read_materials import get_materials_from_bulk
from .read_orientations import get_lcsys_from_bulk
from .read_sections import get_connector_sections_from_bulk, get_sections_from_inp

part_name_counter = Counter(1, "Part")


if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly, Part
    from ada.fem import FEM

logger = get_logger()


@dataclass
class InstanceData:
    part_ref: str
    instance_name: str
    instance_bulk: str
    transform: Transform = field(default_factory=Transform)


def read_fem(fem_file, fem_name=None) -> Assembly:
    """This will create and add an AbaqusPart object based on a path reference to a Abaqus input file."""
    from ada import Assembly

    print("Starting import of Abaqus input file")

    if fem_name is not None:
        global part_name_counter
        part_name_counter = Counter(1, fem_name)

    assembly = Assembly("TempAssembly")

    bulk_str = read_bulk_w_includes(fem_file)
    lbulk = bulk_str.lower()
    ass_start = lbulk.find("\n*assembly")
    ass_end = lbulk.rfind("\n*end assembly")
    step_start = lbulk.rfind("\n*step")

    ass_start = ass_start + 1 if ass_start != -1 else ass_start
    ass_end = ass_end + 1 if ass_end != -1 else ass_end
    step_start = step_start + 1 if step_start != -1 else step_start

    if ass_start == -1 and ass_end == -1:
        uses_assembly_parts = False
        assembly_str = bulk_str
        props_str = bulk_str
    else:
        uses_assembly_parts = True
        assembly_str = bulk_str[ass_start : ass_end + 2]
        props_str = bulk_str[ass_end + 2 : step_start]

    inst_end = assembly_str.lower().rfind("\n*end instance")
    inst_end = inst_end + 15 if inst_end != -1 else inst_end

    get_materials_from_bulk(assembly, props_str)
    get_intprop_from_lines(assembly, props_str)

    ass_data = extract_instance_data(assembly_str[:inst_end])

    part_list = import_parts(bulk_str[:ass_start], ass_data, assembly)
    if len(part_list) == 0:
        add_fem_without_assembly(bulk_str, assembly)

    if uses_assembly_parts is True:
        ass_sets = assembly_str[inst_end:]
        assembly.fem.nodes += get_nodes_from_inp(ass_sets, assembly.fem)
        assembly.fem.lcsys.update(get_lcsys_from_bulk(ass_sets, assembly.fem))
        assembly.fem.connector_sections.update(get_connector_sections_from_bulk(props_str, assembly.fem))
        assembly.fem.elements += get_elem_from_bulk_str(ass_sets, assembly.fem)
        assembly.fem.elements.build_sets()
        assembly.fem.sets += get_sets_from_bulk(ass_sets, assembly.fem)
        assembly.fem.sets.link_data()

        update_connector_data(ass_sets, assembly.fem)
        assembly.fem.surfaces.update(get_surfaces_from_bulk(ass_sets, assembly.fem))

        try:
            assembly.fem.constraints.update(get_constraints_from_inp(ass_sets, assembly.fem))
        except KeyError as e:
            logger.error(e)

        assembly.fem.bcs += get_bcs_from_bulk(props_str, assembly.fem)
        assembly.fem.elements += get_mass_from_bulk(ass_sets, assembly.fem)

    add_interactions_from_bulk_str(props_str, assembly)
    get_initial_conditions_from_lines(assembly, props_str)
    return assembly


def read_bulk_w_includes(inp_path) -> str:
    re_bulk_include = re.compile(r"\*Include,\s*input=(.*?)$", _re_in)
    bulk_repl = dict()
    with open(inp_path, "r") as inpDeck:
        bulk_str = inpDeck.read()
        for m in re_bulk_include.finditer(bulk_str):
            search_key = m.group(0)
            with open(os.path.join(os.path.dirname(inp_path), m.group(1)), "r") as d:
                bulk_repl[search_key] = d.read()

    for key, val in bulk_repl.items():
        bulk_str = bulk_str.replace(key, val)
    return bulk_str


def import_bulk(file_path, buffer_function):
    with open(file_path, "r") as f:
        return buffer_function(f.read())


def import_bulk2(file_path, buffer_function):
    with open(file_path, "r") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as m:
            return buffer_function(m)


def extract_instance_data(assembly_bulk) -> dict[str, List[InstanceData]]:
    ass_data = {}
    for m in cards.inst_matches.finditer(assembly_bulk):
        d = m.groupdict()
        inst_name, part_name, bulk_str = d["inst_name"], d["part_name"], d["bulk_str"]
        inst_data = get_instance_data(inst_name, part_name, bulk_str)
        if inst_data.part_ref not in ass_data.keys():
            ass_data[inst_data.part_ref] = []
        ass_data[inst_data.part_ref].append(inst_data)

    return ass_data


def import_parts(bulk_str, instance_data: dict[str, List[InstanceData]], assembly: Assembly) -> List[Part]:
    part_list = []

    for m in cards.parts_matches.finditer(bulk_str):
        d = m.groupdict()
        name = d.get("name")
        part_bulk_str = d.get("bulk_str")

        for i in instance_data[name]:
            p_bulk_str = i.instance_bulk if part_bulk_str == "" and i.instance_bulk != "" else part_bulk_str
            part = get_fem_from_bulk_str(name, p_bulk_str, assembly, i)
            part_list.append(part)
    return part_list


def add_fem_without_assembly(bulk_str, assembly: Assembly) -> Part:
    part_name_matches = list(cards.part_names.finditer(bulk_str))
    p_nmatch = tuple(part_name_matches)

    if len(p_nmatch) != 1:
        p_bulk = bulk_str
        p_name = None
    else:
        p_name = p_nmatch[0].group(1)
        p_bulk = p_nmatch[0].group(2)

    p_name = next(part_name_counter) if p_name is None else p_name
    inst = InstanceData("", p_name, "")

    return get_fem_from_bulk_str(p_name, p_bulk, assembly, inst)


def get_fem_from_bulk_str(name, bulk_str, assembly: Assembly, instance_data: InstanceData) -> "Part":
    from ada import FEM, Part

    instance_name = name if instance_data.instance_name is None else instance_data.instance_name
    if name in assembly.parts.keys():
        name = instance_name
    part = assembly.add_part(Part(name, fem=FEM(name=instance_name)))
    fem = part.fem
    fem.nodes = get_nodes_from_inp(bulk_str, fem)
    fem.nodes.move(move=instance_data.transform.translation, rotate=instance_data.transform.rotation)
    fem.elements = get_elem_from_bulk_str(bulk_str, fem)
    fem.elements.build_sets()
    fem.sets += get_sets_from_bulk(bulk_str, fem)
    fem.sections = get_sections_from_inp(bulk_str, fem)
    fem.bcs += get_bcs_from_bulk(bulk_str, fem)
    fem.elements += get_mass_from_bulk(bulk_str, fem)
    fem.surfaces.update(get_surfaces_from_bulk(bulk_str, fem))
    fem.lcsys = get_lcsys_from_bulk(bulk_str, fem)
    fem.constraints = get_constraints_from_inp(bulk_str, fem)

    print(8 * "-" + f'Imported "{part.fem.instance_name}"')

    return part


def get_initial_conditions_from_lines(assembly: Assembly, bulk_str: str):
    """
    TODO: Optimize this function


    ** PREDEFINED FIELDS
    **
    ** Name: IC-1   Type: Velocity
    *Initial Conditions, type=VELOCITY
    CONTAINER20FT-10000KG, 1, 0.
    CONTAINER20FT-10000KG, 2, 0.
    CONTAINER20FT-10000KG, 3, 4.1
    """
    if bulk_str.find("*Initial Conditions") == -1:
        return
    re_str = r"(?:\*\*\s*Name:\s*(.*?)\s*Type:\s*(.*?)\n)+\*Initial Conditions, type=.*?\n(?<=)((?:.*?)(?=\*|\Z))"

    def sort_props(line):
        ev = [x.strip() for x in line.split(",")]
        set_name = ev[0]
        try:
            dofs = int(ev[1])
        except BaseException as e:
            logger.debug(e)
            dofs = ev[1]
        if len(ev) > 2:
            magn = ev[2]
        else:
            magn = None
        return set_name, dofs, magn

    def grab_init_props(m):
        bc_name = m.group(1)
        bc_type = m.group(2)
        bcs = m.group(3)
        props = [sort_props(line) for line in bcs.splitlines()]
        set_name, dofs, magn = list(zip(*props))
        fem_set = None
        set_name_up = set_name[0]
        if set_name_up in assembly.fem.elsets.keys():
            fem_set = assembly.fem.elsets[set_name_up]
        elif set_name_up in assembly.fem.nsets.keys():
            fem_set = assembly.fem.nsets[set_name_up]
        else:
            for p in assembly.get_all_parts_in_assembly():
                if set_name_up in p.fem.elsets.keys():
                    fem_set = p.fem.elsets[set_name_up]
        if fem_set is None:
            raise ValueError(f'Unable to find fem set "{set_name[0]}"')

        return PredefinedField(bc_name, bc_type, fem_set, dofs, magn, parent=assembly.fem)

    for match in re.finditer(re_str, bulk_str, _re_in):
        assembly.fem.add_predefined_field(grab_init_props(match))


def get_intprop_from_lines(assembly: Assembly, bulk_str):
    """
    *Surface Interaction, name=contactProp
    *Friction
    0.,
    *Surface Behavior, pressure-overclosure=HARD
    """
    re_str = r"(\*Surface Interaction, name=.*?)(?=\*)(?!\*Friction|\*Surface Behavior|\*Surface Smoothing)"
    # surf_interact = AbaFF("Surface Interaction", [("name=",), ("bulk>",)], [("Friction", [(), ("bulkFRIC>",)])])

    assembly.fem.metadata["surf_smoothing"] = []
    for m in cards.surface_smoothing.regex.finditer(bulk_str):
        d = m.groupdict()
        assembly.fem.metadata["surf_smoothing"].append(d)

    for m in re.findall(re_str, bulk_str, _re_in):
        name = re.search(r"name=(.*?)\n", m, _re_in).group(1)
        re_fric = re.search(r"\*Friction\n(.*?),", m, _re_in)
        fric = None
        if re_fric is not None:
            fric = re_fric.group(1)
        props = dict(name=name, friction=fric)
        res = re.search(
            r"\*Surface Behavior,\s*(?P<btype>.*?)(?:=(?P<behaviour>.*?)(?:$\n(?P<tabular>.*?)(?:\*|\Z)|$)|$)",
            m,
            _re_in,
        )
        if res is not None:
            res = res.groupdict()
            btype = res["btype"]
            behave = res["behaviour"] if btype.upper() != "PENALTY" else btype
            tabular = res["tabular"]
            if tabular is not None:
                tab_list = []
                for line in tabular.splitlines():
                    tab_list.append(tuple(np.fromstring(line, dtype=float, sep=",")))
                tabular = tab_list
                props.update(dict(pressure_overclosure=behave, tabular=tabular))
        assembly.fem.add_interaction_property(InteractionProperty(**props))


def get_instance_data(inst_name, p_ref, inst_bulk) -> InstanceData:
    """Move/rotate data lines are specified here:

    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-instance.htm
    """

    move_rot = re.compile(r"(?:^\s*(.*?),\s*(.*?),\s*(.*?)$)", _re_in)
    transform: Union[Transform, None] = Transform()
    mr = move_rot.finditer(inst_bulk)
    if mr is not None:
        for j, mo in enumerate(mr):
            content = mo.group(0)
            if "*" in content or j == 2 or content == "":
                break
            if j == 0:
                transform.move = (float(mo.group(1)), float(mo.group(2)), float(mo.group(3)))
            if j == 1:
                r = [float(x) for x in mo.group(3).split(",")]
                origin = (float(mo.group(1)), float(mo.group(2)), r[0])
                vector = (r[1], r[2], r[3])
                angle = r[4]
                transform.rotate = Rotation(origin, vector, angle)

    return InstanceData(p_ref, inst_name, inst_bulk, transform)


def import_multiple_inps(input_files_dir):
    """
    Import a set of inp files from a folder

    :param input_files_dir:
    """

    def read_inp(fname):
        if ".inp" not in fname:
            return None
        else:
            with open(pathlib.Path(input_files_dir) / fname, "r") as d:
                return d.read()

    return "".join([x for x in map(read_inp, os.listdir(input_files_dir)) if x is not None])


def get_nodes_from_inp(bulk_str, parent: FEM) -> Nodes:
    """Extract node information from abaqus input file string"""
    re_no = re.compile(
        r"^\*Node\s*(?:,\s*nset=(?P<nset>.*?)\n|\n)(?P<members>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def getnodes(m):
        d = m.groupdict()
        res = np.fromstring(list_cleanup(d["members"]), sep=",", dtype=np.float64)
        res_ = res.reshape(int(res.size / 4), 4)
        members = [Node(n[1:4], int(n[0]), parent=parent) for n in res_]
        if d["nset"] is not None:
            parent.sets.add(FemSet(d["nset"], members, "nset", parent=parent))
        return members

    nodes = list(chain.from_iterable(map(getnodes, re_no.finditer(bulk_str))))

    return Nodes(nodes, parent=parent)


def str_to_ints(instr):
    try:
        return [int(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]
    except ValueError:
        return [str(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]


def get_sets_from_bulk(bulk_str, fem: FEM) -> FemSets:
    from ada import Assembly

    if fem.parent is not None:
        all_parts = fem.parent.get_all_parts_in_assembly()
    else:
        all_parts = []

    def get_parent_instance(instance) -> FEM:
        if instance is None or fem.parent is None:
            return fem
        elif instance is not None and type(fem.parent) is Assembly:
            for p in filter(lambda x: x.fem.instance_name == instance, all_parts):
                return p.fem
        else:
            raise ValueError(f'Unable to find instance "{instance}" amongst assembly parts')

    def get_set(match):
        name = match.group(2)
        set_type = match.group(1)
        internal = True if match.group(3) is not None else False
        instance = match.group(4)
        generate = True if match.group(5) is not None else False
        members_str = match.group(6)
        gen_mem = str_to_ints(members_str) if generate is True else []
        members = [] if generate is True else str_to_ints(members_str)
        metadata = dict(instance=instance, internal=internal, generate=generate, gen_mem=gen_mem)
        parent_instance = get_parent_instance(instance)

        if set_type.lower() == "elset":
            members = [parent_instance.elements.from_id(el_id) for el_id in members]
        else:
            members = [parent_instance.nodes.from_id(el_id) for el_id in members]

        fem_set = FemSet(
            name,
            members,
            set_type=set_type,
            metadata=metadata,
            parent=parent_instance,
        )

        return fem_set

    return FemSets([get_set(x) for x in cards.re_sets.finditer(bulk_str)], parent=fem)

    # import concurrent.futures
    # with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
    #     futures = []
    #     for number in re_sets.finditer(bulk_str):
    #         future = executor.submit(get_set, number)
    #         futures.append(future)
    #     sets = [future.result() for future in concurrent.futures.as_completed(futures)]
    #
    # return FemSetsCollection(sets, parent)


def get_bcs_from_bulk(bulk_str, fem: FEM) -> List[Bc]:
    bc_counter = Counter(1, "bc")

    def get_dofs(content: str):
        set_name = None
        dofs = []
        magn = []
        if content.count("\n") == 1:
            temp = content.split(",")
            set_name = temp[0].strip()
            dof_in = temp[1].strip().replace("\n", "")
            if dof_in.lower() == "encastre":
                dofs = dof_in
            else:
                dof = str_to_int(temp[1])
                if len(temp) == 2:
                    dofs = [x if dof == x else None for x in range(1, 7)]
                else:
                    dof_end = str_to_int(temp[2])
                    dofs = [x if dof <= x <= dof_end else None for x in range(1, 7)]

        else:
            for line in content.splitlines():
                ev = [x.strip() for x in line.split(",")]
                set_name = ev[0]
                try:
                    dofs.append(int(ev[1]))
                except BaseException as e:
                    logger.debug(e)
                    dofs.append(ev[1])
                if len(ev) > 3:
                    magn.append(ev[2])
        magn = None if len(magn) == 0 else magn
        return set_name, dofs, magn

    def get_nset(part_instance_name, set_name):
        for p in fem.parent.parts.values():
            if p.fem.instance_name == part_instance_name:
                return p.fem.sets.get_nset_from_name(set_name)

    def get_bc(match):
        d = match.groupdict()
        bc_name = d["name"] if d["name"] is not None else next(bc_counter)
        bc_type = d["type"]
        set_name, dofs, magn = get_dofs(d["content"])

        if "." in set_name:
            part_instance_name, set_name = set_name.split(".")
            fem_set = get_nset(part_instance_name, set_name)
        else:
            if set_name in fem.nsets.keys():
                fem_set = fem.sets.get_nset_from_name(set_name)
            else:
                val = str_to_int(set_name)
                if val in fem.nodes.dmap.keys():
                    node = fem.nodes.from_id(val)
                    fem_set = FemSet(bc_name + "_set", [node], "nset", parent=fem)
                else:
                    raise ValueError(f'Unable to find set "{set_name}" in part {fem}')

        if fem_set is None:
            raise Exception("Unable to Find node set")

        props = dict()
        if bc_type is not None:
            props["bc_type"] = bc_type
        if magn is not None:
            props["magnitudes"] = magn

        return Bc(bc_name, fem_set, dofs, parent=fem, **props)

    return [get_bc(match_in) for match_in in cards.re_bcs.finditer(bulk_str)]


def get_surfaces_from_bulk(bulk_str, parent):
    from ada.fem.elements import find_element_type_from_list
    from ada.fem.surfaces import SurfTypes

    def interpret_member(mem):
        msplit = mem.split(",")
        try:
            ref = str_to_int(msplit[0])
        except BaseException as e:
            logger.debug(e)
            ref = msplit[0].strip()

        return tuple([ref, msplit[1].strip()])

    surf_d = dict()

    for m in cards.surface.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].strip()
        surf_type = d["type"].upper() if d["type"] is not None else "ELEMENT"
        members_str: str = d["bulk"]
        if members_str.count("\n") >= 1:
            id_refs = [interpret_member(m) for m in members_str.splitlines()]
            set_ref, set_id_ref = None, None
        else:
            id_refs = None
            res = [x.strip() for x in members_str.split(",")]
            if len(res) == 2:
                set_ref, set_id_ref = res
            else:
                set_ref = res[0]
                set_id_ref = 1.0

        if id_refs is None:
            if surf_type == SurfTypes.NODE:
                if set_id_ref == "":
                    fem_set = FemSet(f"n{set_ref}_set", [parent.nodes.from_id(int(set_ref))], "nset")
                    parent.add_set(fem_set)
                    weight_factor = 1.0
                else:
                    if "." in set_ref:
                        ssplit = set_ref.split(".")
                        parent_ = None
                        fem_set = None
                        for prt in parent.parent.get_all_parts_in_assembly():
                            if prt.fem.name == ssplit[0]:
                                parent_ = prt.fem
                                fem_set = prt.fem.nsets[ssplit[1]]
                                break
                        if parent_ is None:
                            raise ValueError(f'Unable to find parent FEM "{ssplit[0]}"')
                    else:
                        fem_set = parent.nsets[set_ref]
                    weight_factor = float(set_id_ref)
                el_face_index = None
            else:
                weight_factor = None
                fem_set = parent.sets.get_elset_from_name(set_ref)
                el_type = find_element_type_from_list(fem_set.members)
                if el_type == ElemType.SOLID:
                    el_face_index = int(set_id_ref.replace("S", "")) - 1
                elif el_type == ElemType.SHELL:
                    el_face_index = -1 if set_id_ref == "SNEG" else 1
                else:
                    el_face_index = set_id_ref
        else:
            fem_set = None
            weight_factor = None
            el_face_index = None

        surf_d[name] = Surface(
            name,
            surf_type,
            fem_set,
            weight_factor,
            el_face_index,
            id_refs,
            parent=parent,
        )

    return surf_d


def get_constraints_from_inp(bulk_str: str, fem: FEM) -> Dict[str, Constraint]:
    """

    ** Constraint: Container_RigidBody
    *Rigid Body, ref node=container_rp, elset=container

    *MPC
     BEAM,    2007,     161
     BEAM,    2008,     162
    """

    # Rigid Bodies

    constraints = []
    rbnames = Counter(1, "rgb")
    conames = Counter(1, "co")

    for m in cards.tie.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        msurf = get_set_from_assembly(d["surf1"], fem, "surface")
        ssurf = get_set_from_assembly(d["surf2"], fem, "surface")
        constraints.append(Constraint(name, Constraint.TYPES.TIE, msurf, ssurf, metadata=dict(adjust=d["adjust"])))

    for m in cards.rigid_bodies.regex.finditer(bulk_str):
        d = m.groupdict()
        name = next(rbnames)
        ref_node = get_set_from_assembly(d["ref_node"], fem, FemSet.TYPES.NSET)
        elset = get_set_from_assembly(d["elset"], fem, FemSet.TYPES.ELSET)
        constraints.append(Constraint(name, Constraint.TYPES.RIGID_BODY, ref_node, elset, parent=fem))

    couplings = []
    for m in cards.coupling.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        rn = d["ref_node"].strip()
        sf = d["surface"].strip()
        if rn.isnumeric():
            ref_set = FemSet(next(conames), [fem.nodes.from_id(int(rn))], FemSet.TYPES.NSET, parent=fem)
            fem.sets.add(ref_set)
        else:
            ref_set = fem.nsets[rn]

        surf = fem.surfaces[sf]

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=int)
        size = res.size
        cols = 2
        rows = int(size / cols)
        dofs = res.reshape(rows, cols)

        csys_name = d.get("orientation", None)
        if csys_name is not None:
            if csys_name not in fem.lcsys.keys():
                raise ValueError(f'Csys "{csys_name}" was not found on part {fem}')
            csys = fem.lcsys[csys_name]
        else:
            csys = None

        couplings.append(Constraint(name, Constraint.TYPES.COUPLING, ref_set, surf, csys=csys, dofs=dofs, parent=fem))

    # Shell to Solid Couplings
    sh2solids = []
    for m in cards.sh2so_re.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        influence = d["influence_distance"]
        surf1 = get_set_from_assembly(d["surf1"], fem, "surface")
        surf2 = get_set_from_assembly(d["surf2"], fem, "surface")
        sh2solids.append(Constraint(name, Constraint.TYPES.SHELL2SOLID, surf1, surf2, influence_distance=influence))

    # MPC's
    mpc_dict = dict()
    mpc_re = re.compile(r"\*mpc\n(?P<bulk>.*?)^\*", _re_in)
    mpc_content = re.compile(r"\s*(?P<mpc_type>.*?),\s*(?P<master>.*?),\s*(?P<slave>.*?)$", _re_in)
    for match in mpc_re.finditer(bulk_str):
        d1 = match.groupdict()
        for subm in mpc_content.finditer(d1["bulk"]):
            d = subm.groupdict()
            mpc_type = d["mpc_type"]
            m = d["master"]
            s = d["slave"]
            if mpc_type not in mpc_dict.keys():
                mpc_dict[mpc_type] = []
            try:
                n1_ = str_to_int(m)
            except BaseException as e:
                logger.debug(e)
                n1_ = get_set_from_assembly(m, fem, FemSet.TYPES.NSET)

            try:
                n2_ = str_to_int(s)
            except BaseException as e:
                logger.debug(e)
                n2_ = get_set_from_assembly(s, fem, FemSet.TYPES.NSET)

            mpc_dict[mpc_type].append((n1_, n2_))

    def get_mpc(mpc_values):
        m_set, s_set = zip(*mpc_values)
        mpc_name = mpc_type + "_mpc"
        mset = FemSet("mpc_" + mpc_type + "_m", m_set, FemSet.TYPES.NSET)
        sset = FemSet("mpc_" + mpc_type + "_s", s_set, FemSet.TYPES.NSET)
        return Constraint(mpc_name, Constraint.TYPES.MPC, mset, sset, mpc_type=mpc_type, parent=fem)

    mpcs = [get_mpc(mpc_values_in) for mpc_values_in in mpc_dict.values()]

    return {c.name: c for c in chain.from_iterable([constraints, couplings, sh2solids, mpcs])}


def add_interactions_from_bulk_str(bulk_str, assembly: Assembly) -> None:
    gen_name = Counter(1, "general")

    if bulk_str.find("** Interaction") == -1 and bulk_str.find("*Contact") == -1:
        return

    def resolve_surface_ref(surf_ref):
        surf_name = surf_ref.split(".")[-1] if "." in surf_ref else surf_ref
        surf = None
        if surf_name in assembly.fem.surfaces.keys():
            surf = assembly.fem.surfaces[surf_name]

        for p in assembly.get_all_parts_in_assembly():
            if surf_name in p.fem.surfaces.keys():
                surf = p.fem.surfaces[surf_name]
        if surf is None:
            raise ValueError("Unable to find surfaces in assembly parts")

        return surf

    for m in cards.contact_pairs.regex.finditer(bulk_str):
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        surf1 = resolve_surface_ref(d["surf1"])
        surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(Interaction(d["name"], ContactTypes.SURFACE, surf1, surf2, intprop, metadata=d))

    for m in cards.contact_general.regex.finditer(bulk_str):
        s = m.start()
        e = m.endpos
        interact_str = bulk_str[s:e]
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        # surf1 = resolve_surface_ref(d["surf1"])
        # surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(
            Interaction(next(gen_name), "general", None, None, intprop, metadata=dict(aba_bulk=interact_str))
        )



from ada.config import get_logger

from .exceptions import InsufficientPointsError

logger = get_logger()


def get_plates_from_satd(satd: dict) -> dict:
    plate_geom = dict()

    for face in filter(lambda x: x[0] == "face", satd.values()):
        try:
            res = get_face_from_satd(face, satd)
        except KeyError as e:
            logger.warning(f"Unable to import face due to KeyError -> {e}")
            continue
        except InsufficientPointsError as e:
            logger.warning(f"Unable to import face due to Insufficient points -> {e}")
            continue
        plate_geom.update(res)

    return plate_geom


def get_face_from_satd(face: tuple, satd: dict):
    # Face row
    name_idx = 1
    loop_idx = 6

    # Loop row
    coedge_ref = 6

    face_ref = get_value_from_satd(face[name_idx], satd)[-1]

    loop = get_value_from_satd(face[loop_idx], satd)
    coedge_start_id = loop[coedge_ref]
    coedge_first = get_value_from_satd(coedge_start_id, satd)

    coedge_first_direction = str(coedge_first[-3])

    # Coedge row
    next_coedge_idx = 5 if coedge_first_direction == "forward" else 6

    next_coedge = True
    coedge_next_id = coedge_first[next_coedge_idx]
    edges = [coedge_first]

    max_iter = 500
    i = 0
    while next_coedge is True:
        coedge = get_value_from_satd(coedge_next_id, satd)
        edges.append(coedge)

        coedge_next_id = coedge[next_coedge_idx]
        if coedge_next_id == coedge_start_id:
            next_coedge = False

        i += 1
        if i > max_iter:
            raise ValueError(f"Found {i} points which is over max={max_iter}")

    p1, p2 = get_points_from_edge(coedge_first, satd)

    points = [p1, p2]

    for coedge in edges:
        p1, p2 = get_points_from_edge(coedge, satd)
        edge_direction = str(coedge[-3])
        if edge_direction == "forward":
            p = p2
        else:
            p = p1
        if p not in points:
            points.append(p)

    if len(points) < 3:
        raise InsufficientPointsError("Plates cannot have < 3 points")

    if coedge_first_direction == "reversed":
        points.reverse()

    return {face_ref: points}


def get_points_from_edge(coedge: tuple, satd: dict):
    # Coedge row
    edge_ref = 8

    # Edge row
    vert1_idx = 5
    vert2_idx = 7
    # edge_type_idx = 8

    # Vertex row
    p_idx = -1

    edge = get_value_from_satd(coedge[edge_ref], satd)
    vert1 = get_value_from_satd(edge[vert1_idx], satd)
    vert2 = get_value_from_satd(edge[vert2_idx], satd)
    # edge_type = get_value_from_satd(edge[edge_type_idx], satd)
    p1 = get_value_from_satd(vert1[p_idx], satd)
    p2 = get_value_from_satd(vert2[p_idx], satd)
    n1 = tuple([float(x) for x in p1[-3:]])
    n2 = tuple([float(x) for x in p2[-3:]])
    return n1, n2


def get_value_from_satd(val_str: str, satd: dict) -> tuple:
    return satd[val_str.replace("$", "")]



from __future__ import annotations

import os
import pathlib
from dataclasses import dataclass, field
from typing import Iterable, List, Union

import gmsh
import numpy as np

from ada import FEM, Beam, Pipe, Plate, Shape
from ada.base.physical_objects import BackendGeom
from ada.base.types import GeomRepr
from ada.concepts.containers import Nodes
from ada.config import Settings, get_logger
from ada.fem import Elem
from ada.fem.containers import FemElements
from ada.fem.shapes import ElemType
from ada.ifc.utils import create_guid

logger = get_logger()


@dataclass
class GmshOptions:
    # Mesh
    Mesh_Algorithm: int = 6  #
    Mesh_ElementOrder: int = 1
    Mesh_Algorithm3D: int = 1
    Mesh_MeshSizeMin: float = None
    Mesh_MeshSizeMax: float = 0.1
    Mesh_SecondOrderIncomplete: int = 1
    Mesh_Smoothing: int = 3
    Mesh_RecombinationAlgorithm: int = None
    Mesh_SubdivisionAlgorithm: int = None
    # Curvature Options
    Mesh_MeshSizeFromCurvature: bool = False
    Mesh_MinimumElementsPerTwoPi: int = 12
    # Geometry
    Geometry_Tolerance: float = 1e-5
    Geometry_OCCImportLabels: int = 1
    Geometry_OCCMakeSolids: int = None
    # General (UI)
    General_ColorScheme: int = 3
    General_Orthographic: int = 0  # 0 Means perspective projection
    General_Terminal: int = 1

    def get_as_dict(self):
        return {key.replace("_", "."): value for key, value in vars(self).items() if value is not None}


@dataclass
class GmshTask:
    ada_obj: List[Union[Shape, Beam, Plate]]
    geom_repr: str
    mesh_size: float
    options: GmshOptions = field(default_factory=GmshOptions)


@dataclass
class CutPlane:
    origin: tuple
    dx: float = 10
    dy: float = 10
    plane: str = "XY"
    normal: tuple = None
    cut_objects: List[GmshData] = None
    gmsh_id: int = None


@dataclass
class GmshData:
    entities: Iterable
    geom_repr: str
    order: int
    obj: Shape | Beam | Plate | Pipe
    mesh_size: float = None


class GmshSession:
    def __init__(self, silent=False, persist=True, options: GmshOptions = GmshOptions()):
        logger.debug("init method called")
        self._gmsh = None
        self.options = options
        self.cutting_planes: List[CutPlane] = []
        self.silent = silent
        if silent is True:
            self.options.General_Terminal = 0
        self.persist = persist
        self.model_map: dict[Union[Shape, Beam, Plate, Pipe], GmshData] = dict()

    def add_obj(
        self,
        obj: BackendGeom | Shape | Beam | Plate | Pipe,
        geom_repr: GeomRepr | str = ElemType.SOLID,
        el_order=1,
        silent=True,
        mesh_size=None,
        build_native_lines=False,
        point_tol=Settings.point_tol,
        use_native_pointer=True,
    ):
        from ada.core.utils import Counter

        from .utils import build_bm_lines

        if isinstance(geom_repr, str):
            geom_repr = GeomRepr.from_str(geom_repr)

        self.apply_settings()
        temp_dir = Settings.temp_dir
        os.makedirs(temp_dir, exist_ok=True)

        if build_native_lines is True and geom_repr == ElemType.LINE and type(obj) is Beam:
            entities = build_bm_lines(self.model, obj, point_tol)
        else:
            if use_native_pointer and hasattr(self.model.occ, "importShapesNativePointer"):
                # Use hasattr to ensure that it works for gmsh < 4.9.*
                if type(obj) is Pipe:
                    entities = []
                    for seg in obj.segments:
                        entities += import_into_gmsh_use_nativepointer(seg, geom_repr, self.model)
                else:
                    entities = import_into_gmsh_use_nativepointer(obj, geom_repr, self.model)
            else:
                entities = import_into_gmsh_using_step(obj, geom_repr, self.model, temp_dir, silent)

        obj_name = Counter(1, f"{obj.name}_")
        for dim, ent in entities:
            ent_name = next(obj_name)
            self.model.set_physical_name(dim, ent, ent_name)
            self.model.set_entity_name(dim, ent, ent_name)

        self.model.occ.synchronize()
        self.model.geo.synchronize()

        gmsh_data = GmshData(entities, geom_repr, el_order, obj, mesh_size=mesh_size)
        self.model_map[obj] = gmsh_data
        return gmsh_data

    def add_cutting_plane(self, cut_plane: CutPlane, cut_objects: List[GmshData] = None):
        if cut_objects is not None:
            if cut_plane.cut_objects is None:
                cut_plane.cut_objects = []
            cut_plane.cut_objects += cut_objects
        self.cutting_planes.append(cut_plane)

    def make_cuts(self):
        geom_repr_map = {ElemType.SOLID: 3, ElemType.SHELL: 2, ElemType.LINE: 1}

        for cut in self.cutting_planes:
            x, y, z = cut.origin
            rect = self.model.occ.addRectangle(x, y, z, cut.dx, cut.dy)

            if cut.plane == "XZ":
                self.model.occ.synchronize()
                self.model.geo.synchronize()
                self.model.occ.rotate([(2, rect)], x, y, z, 1, 0, 0, np.deg2rad(90))
                self.model.occ.synchronize()
                self.model.geo.synchronize()

            cut.gmsh_id = rect
            for obj in cut.cut_objects:
                res, _ = self.model.occ.fragment(obj.entities, [(2, rect)], removeTool=True)
                cut_geom_dim = geom_repr_map[obj.geom_repr]
                replaced_entities = [(dim, r) for dim, r in res if r != rect and dim == cut_geom_dim]
                obj.entities = replaced_entities
            self.model.occ.remove([(2, rect)], True)

        # rem_ids = [(2, c.gmsh_id) for c in self.cutting_planes]
        # self.model.occ.remove(rem_ids, True)

        self.model.occ.synchronize()
        self.model.geo.synchronize()

    def split_crossing_beams(self):
        # Todo: base this algo on beams that are actually clashing

        beams = [obj for obj in self.model_map.keys() if type(obj) is Beam]
        if len(beams) == 1:
            return None

        intersecting_beams = []
        int_bm_map = dict()
        for bm in beams:
            bm_gmsh_obj = self.model_map[bm]
            for li_dim, li_ent in bm_gmsh_obj.entities:
                intersecting_beams.append((li_dim, li_ent))
                int_bm_map[(li_dim, li_ent)] = bm_gmsh_obj

        res, res_map = self.model.occ.fragment(intersecting_beams, intersecting_beams)

        for i, int_bm in enumerate(intersecting_beams):
            bm_gmsh_obj = int_bm_map[int_bm]
            new_ents = res_map[i]
            bm_gmsh_obj.entities = new_ents

        self.model.occ.synchronize()

    def split_plates_by_beams(self):
        from ada.core.clash_check import (
            filter_away_beams_along_plate_edges,
            find_beams_connected_to_plate,
        )

        beams = [obj for obj in self.model_map.keys() if type(obj) is Beam]
        plates = [obj for obj in self.model_map.keys() if type(obj) is Plate]
        for pl in plates:
            pl_gmsh_obj = self.model_map[pl]
            for pl_dim, pl_ent in pl_gmsh_obj.entities:
                intersecting_beams = []
                int_bm_map = dict()
                all_contained_beams = find_beams_connected_to_plate(pl, beams)
                inside_beams = filter_away_beams_along_plate_edges(pl, all_contained_beams)
                for bm in inside_beams:
                    bm_gmsh_obj = self.model_map[bm]
                    for li_dim, li_ent in bm_gmsh_obj.entities:
                        intersecting_beams.append((li_dim, li_ent))
                        int_bm_map[(li_dim, li_ent)] = bm_gmsh_obj
                # Using Embed fails during meshing

                # res = self.model.mesh.embed(1, [t for e,t in intersecting_beams], 2, pl_ent)

                res, res_map = self.model.occ.fragment(intersecting_beams, [(pl_dim, pl_ent)])
                replaced_pl_entities = [(dim, r) for dim, r in res if dim == 2]
                for i, int_bm in enumerate(intersecting_beams):
                    bm_gmsh_obj = int_bm_map[int_bm]
                    new_ents = res_map[i]
                    bm_gmsh_obj.entities = new_ents
                pl_gmsh_obj.entities = replaced_pl_entities

                self.model.occ.synchronize()

    def mesh(self, size: float = None, use_quads=False, use_hex=False):
        if self.silent is True:
            self.options.General_Terminal = 0

        if use_quads:
            self.make_quads()

        if use_hex:
            self.make_hex()

        if size is not None:
            self.options.Mesh_MeshSizeMax = size
            if self.options.Mesh_MeshSizeFromCurvature is False:
                self.options.Mesh_MeshSizeMin = size

        self.apply_settings()
        self.model.geo.synchronize()
        self.model.mesh.setRecombine(3, -1)
        self.model.mesh.generate(3)
        self.model.mesh.removeDuplicateNodes()

        if use_quads is True or use_hex is True:
            self.model.mesh.recombine()

    def make_quads(self):
        from ada.fem.meshing.partitioning.strategies import partition_objects_with_holes

        ents = []
        for obj, model in self.model_map.items():
            if model.geom_repr == ElemType.SHELL:
                if len(obj.penetrations) > 0:
                    partition_objects_with_holes(model, self)
                else:
                    for dim, tag in model.entities:
                        ents.append(tag)
                        self.model.mesh.set_transfinite_surface(tag)
                        self.model.mesh.setRecombine(dim, tag)

    def make_hex(self):
        from ada.fem.meshing.partitioning.strategies import partition_solid_beams

        for dim, tag in self.model.get_entities():
            if dim == 2:
                self.model.mesh.set_transfinite_surface(tag)
                self.model.mesh.setRecombine(dim, tag)

        for obj, model in self.model_map.items():
            if model.geom_repr == GeomRepr.SOLID:
                if isinstance(obj, Beam):
                    partition_solid_beams(model, self)
                else:
                    for dim, tag in model.entities:
                        self.model.mesh.set_transfinite_volume(tag)
                        self.model.mesh.setRecombine(dim, tag)

        self.model.mesh.recombine()

    def get_fem(self, name="AdaFEM") -> FEM:
        from .utils import (
            add_fem_sections,
            get_elements_from_entities,
            get_nodes_from_gmsh,
        )

        fem = FEM(name)
        gmsh_nodes = get_nodes_from_gmsh(self.model, fem)
        fem.nodes = Nodes(gmsh_nodes, parent=fem)

        def add_obj_to_elem_ref(el: Elem, obj: Shape | Beam | Plate | Pipe):
            el.refs.append(obj)

        # Get Elements
        elements = []
        for gmsh_data in self.model_map.values():
            entity_elements = get_elements_from_entities(self.model, gmsh_data.entities, fem)
            gmsh_data.obj.elem_refs = entity_elements
            [add_obj_to_elem_ref(el, gmsh_data.obj) for el in entity_elements]
            elements += entity_elements

        fem.elements = FemElements(elements, fem_obj=fem)

        # Add FEM sections
        for model_obj, gmsh_data in self.model_map.items():
            add_fem_sections(self.model, fem, model_obj, gmsh_data)

        fem.nodes.renumber()
        fem.elements.renumber()
        return fem

    def apply_settings(self):
        if self.options is not None:
            for setting, value in self.options.get_as_dict().items():
                self.gmsh.option.setNumber(setting, value)
        self.model.geo.synchronize()

    def open_gui(self):
        self.gmsh.fltk.run()

    def __enter__(self):
        logger.debug("Starting GMSH session")
        self._gmsh = gmsh
        self.gmsh.initialize()
        # self.model.add("ada")
        self.apply_settings()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        logger.debug("Closing GMSH")
        self.gmsh.finalize()

    @property
    def gmsh(self) -> "gmsh":
        return self._gmsh

    @property
    def model(self) -> "gmsh.model":
        return self.gmsh.model


def import_into_gmsh_using_step(
    obj, geom_repr: GeomRepr, model: gmsh.model, temp_dir: pathlib.Path, silent: bool
) -> List[tuple]:
    name = f"{obj.name}_{create_guid()}"
    obj.to_stp(temp_dir / name, geom_repr=geom_repr, silent=silent, fuse_piping=True)
    ents = model.occ.importShapes(str(temp_dir / f"{name}.stp"))
    return ents


def import_into_gmsh_use_nativepointer(obj: BackendGeom | Shape, geom_repr: GeomRepr, model: gmsh.model) -> List[tuple]:
    from OCC.Extend.TopologyUtils import TopologyExplorer

    from ada import PrimBox

    ents = []
    if geom_repr == GeomRepr.SOLID:
        geom = obj.solid()
        t = TopologyExplorer(geom)
        geom_iter = t.solids()
    elif geom_repr == GeomRepr.SHELL:
        geom = obj.shell() if type(obj) not in (PrimBox,) else obj.geom()
        t = TopologyExplorer(geom)
        geom_iter = t.faces()
    else:
        geom = obj.line()
        t = TopologyExplorer(geom)
        geom_iter = t.edges()

    for shp in geom_iter:
        ents += model.occ.importShapesNativePointer(int(shp.this))

    if len(ents) == 0:
        raise ValueError("No entities found")

    return ents



from typing import Iterable

from ada.fem import FemSection
from ada.fem.containers import FemSections


def create_sections_str(fem_sections: FemSections) -> str:
    mat_assign_str = ""

    beam_sections_str = "\n        POUTRE=(),"
    shell_sections_str = "\n        COQUE=(),"

    if len(fem_sections.shells) > 0:
        mat_assign_str_, shell_sections_str = [
            "".join(x) for x in zip(*[write_shell_section(sh) for sh in fem_sections.shells])
        ]
        mat_assign_str += mat_assign_str_
        shell_sections_str = f"\n        COQUE=(\n{shell_sections_str}\n        ),"

    if len(fem_sections.lines) > 0:
        mat_assign_str_, beam_sections_str, orientations_str = [
            "".join(x) for x in zip(*[write_beam_section(bm) for bm in fem_sections.lines])
        ]
        mat_assign_str += mat_assign_str_
        beam_sections_str = f"\n        POUTRE=(\n{beam_sections_str}\n        ),"
        beam_sections_str += f"\n        ORIENTATION=(\n{orientations_str}\n        ),"

    if len(fem_sections.solids) > 0:
        mat_assign_str += write_solid_section(fem_sections.solids)

    sec_str = ""
    if len(fem_sections.lines) > 0 or len(fem_sections.shells) > 0:
        sec_str = f"""element = AFFE_CARA_ELEM(\n
    MODELE=model,{shell_sections_str}{beam_sections_str}
)"""

    return f"""
material = AFFE_MATERIAU(
    MODELE=model,
    AFFE=(
{mat_assign_str}
    )
)

# Shell elements:
#   EPAIS: thickness
#   VECTEUR: a direction of reference in the tangent plan

{sec_str}
"""


def write_shell_section(fem_sec: FemSection) -> tuple[str, str]:
    mat_name = fem_sec.material.name
    sec_name = fem_sec.elset.name
    #
    local_vec = str(tuple(fem_sec.local_y))
    mat_ = f'		_F(MATER=({mat_name},), GROUP_MA="{sec_name}"),\n'
    sec_str = f"""            _F(
                GROUP_MA=("{sec_name}"),
                EPAIS={fem_sec.thickness},
                VECTEUR={local_vec},
            ),
"""
    return mat_, sec_str


def write_beam_section(fem_sec: FemSection) -> tuple[str, str, str]:
    mat_name = fem_sec.material.name
    sec_name = fem_sec.elset.name
    p = fem_sec.section.properties

    values = ",".join([str(x) for x in [p.Ax, p.Iy, p.Iz, p.Ix]])

    local_vec = str(tuple(fem_sec.local_y))

    mat_ = f'		_F(MATER=({mat_name},), GROUP_MA="{sec_name}"),\n'
    sec_str = f"""            _F(
                GROUP_MA=("{sec_name}"),
                SECTION = 'GENERALE',
                CARA = ('A', 'IY', 'IZ', 'JX'),
                VALE = ({values})
            ),
"""
    orientations = f"""            _F(
                GROUP_MA = '{sec_name}',
                CARA = 'VECT_Y',
                VALE = {local_vec}
            ),
"""

    return mat_, sec_str, orientations


def write_solid_section(fem_sections: Iterable[FemSection]) -> str:
    mat_ = ""
    for fsec in fem_sections:
        mat_ += f'		_F(MATER=({fsec.material.name},), GROUP_MA="{fsec.elset.name}"),\n'
    return mat_



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.ifc.utils import create_axis, create_property_set
from ada.ifc.write.write_shapes import generate_ifc_prim_extrude_geom

if TYPE_CHECKING:
    import ifcopenshell

    from ada.concepts.fasteners import Weld


# https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/IfcFastener.htm
# https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/Pset_FastenerWeld.htm


def write_ifc_fastener(weld: Weld) -> ifcopenshell.entity_instance:
    if weld.parent is None:
        raise ValueError("Parent cannot be None for IFC export")

    a = weld.parent.get_assembly()
    ifc_store = a.ifc_store
    f = a.ifc_store.f

    axis = create_axis(f, [weld.p1.p, weld.p2.p], ifc_store.get_context("Axis"))
    geom = generate_ifc_prim_extrude_geom(weld.geometry, f)
    body = f.createIfcShapeRepresentation(ifc_store.get_context("Body"), "Body", "SweptSolid", [geom])
    shape = f.create_entity("IfcProductDefinitionShape", Name=None, Description=None, Representations=[axis, body])

    ifc_fastener = f.create_entity(
        "IfcFastener",
        GlobalId=weld.guid,
        OwnerHistory=ifc_store.owner_history,
        Name=weld.name,
        Description=None,
        Representation=shape,
        PredefinedType="WELD",
    )

    # https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/Pset_FastenerWeld.htm
    create_property_set("Pset_FastenerWeld", f, dict(Type1=weld.type.value), ifc_store.owner_history)

    return ifc_fastener



from __future__ import annotations

import datetime
import os
import pathlib
from dataclasses import dataclass, field

import numpy as np
import trimesh

from ada.config import get_logger

from .colors import VisColor

logger = get_logger()


@dataclass
class VisMesh:
    """Visual Mesh"""

    name: str
    project: str = None
    world: list[PartMesh] = field(default_factory=list, repr=False)
    meshes: dict[str, VisNode] = field(default_factory=dict, repr=False)
    meta: None | dict = field(default=None, repr=False)
    created: str = None
    translation: np.ndarray = None
    cache_file: pathlib.Path = field(default=pathlib.Path(".cache/meshes.h5"), repr=False)
    overwrite_cache: bool = False
    colors: dict[str, VisColor] = field(default_factory=dict)
    merged: bool = False

    def __post_init__(self):
        if self.created is None:
            self.created = datetime.datetime.utcnow().strftime("%m/%d/%Y, %H:%M:%S")

    def move_objects_to_center(self, override_center=None):
        self.translation = override_center if override_center is not None else -self.vol_center
        for pm in self.world:
            pm.move_objects_to_center(self.translation)

    def add_mesh(self, guid, parent_guid, position, indices, normals=None, matrix=None, color_ref=None):
        obj_group = self._h5cache_group.create_group(guid)
        obj_group.attrs.create("COLOR", color_ref)
        if matrix is not None:
            obj_group.attrs.create("MATRIX", matrix)
        obj_group.create_dataset("POSITION", data=position)
        obj_group.create_dataset("NORMAL", data=normals)
        obj_group.create_dataset("INDEX", data=indices)
        self.meshes[guid] = VisNode(guid, parent_guid)

    @property
    def vol_center(self) -> np.ndarray:
        return (self.bbox[0] + self.bbox[1]) / 2

    @property
    def bbox(self) -> tuple[np.ndarray, np.ndarray]:
        res = np.concatenate([np.array(x.bbox) for x in self.world])
        return res.min(0), res.max(0)

    @property
    def num_polygons(self):
        return sum([x.num_polygons for x in self.world])

    def _convert_to_trimesh(self, embed_meta=True) -> trimesh.Scene:
        scene = trimesh.Scene()
        meta_set = set(self.meta.keys())

        id_sequence = dict()

        for world in self.world:
            world_map_set = set(world.id_map.keys())
            res = meta_set - world_map_set
            if self.merged is False:
                for spatial_node in res:
                    spatial_name, spatial_parent = self.meta.get(spatial_node)
                    scene.graph.update(
                        frame_to=spatial_name, frame_from=spatial_parent if spatial_parent != "*" else None
                    )

            for key, obj in world.id_map.items():
                if self.merged is False:
                    name, parent_guid = self.meta.get(key)
                    parent_name, _ = self.meta.get(parent_guid)
                else:
                    name = key
                    parent_name = "world"

                for i, new_mesh in enumerate(obj.to_trimesh()):
                    name = name if i == 0 else f"{name}_{i:02d}"
                    scene.add_geometry(new_mesh, node_name=name, geom_name=name, parent_node_name=parent_name)
                    id_sequence[name] = obj.id_sequence

        if embed_meta:
            scene.metadata["meta"] = self.meta
            scene.metadata["id_sequence"] = id_sequence

        return scene

    def _export_using_trimesh(self, mesh: trimesh.Scene, dest_file: pathlib.Path):
        os.makedirs(dest_file.parent, exist_ok=True)
        print(f'Writing Visual Mesh to "{dest_file}"')
        with open(dest_file, "wb") as f:
            mesh.export(file_obj=f, file_type=dest_file.suffix[1:])

    def to_stl(self, dest_file):
        dest_file = pathlib.Path(dest_file).with_suffix(".stl")
        mesh: trimesh.Trimesh = self._convert_to_trimesh()
        self._export_using_trimesh(mesh, dest_file)

    def to_gltf(self, dest_file, only_these_guids: list[str] = None, embed_meta=False):
        from ada.core.vector_utils import rot_matrix

        dest_file = pathlib.Path(dest_file).with_suffix(".glb")
        mesh: trimesh.Trimesh = self._convert_to_trimesh(embed_meta=embed_meta)

        # Trimesh automatically transforms by setting up = Y. This will counteract that transform
        m3x3 = rot_matrix((0, -1, 0))
        m3x3_with_col = np.append(m3x3, np.array([[0], [0], [0]]), axis=1)
        m4x4 = np.r_[m3x3_with_col, [np.array([0, 0, 0, 1])]]
        mesh.apply_transform(m4x4)

        self._export_using_trimesh(mesh, dest_file)

    def merge_objects_in_parts_by_color(self) -> VisMesh:
        to_be_merged_part = None
        for pmesh in self.world:
            if to_be_merged_part is None:
                to_be_merged_part = pmesh
                continue
            to_be_merged_part += pmesh
        if to_be_merged_part is None:
            logger.error(f"{self.name} has no parts!?. returning empty model")
            merged_part = []
        else:
            merged_part = to_be_merged_part.merge_by_color()

        return VisMesh(
            name=self.name,
            created=self.created,
            project=self.project,
            world=[merged_part],
            meta=self.meta,
            translation=self.translation,
            merged=True,
        )

    def __add__(self, other: VisMesh):
        new_meta = dict()
        if self.meta is not None:
            new_meta.update(self.meta)
        if other.meta is not None:
            new_meta.update(other.meta)
        return VisMesh(
            name=self.name,
            project=self.project,
            world=self.world + other.world,
            meta=new_meta,
        )


@dataclass
class PartMesh:
    name: str
    id_map: dict[str, ObjectMesh]

    def move_objects_to_center(self, override_center=None):
        for omesh in self.id_map.values():
            oc = override_center if override_center is not None else self.vol_center
            omesh.translate(oc)

    @property
    def vol_center(self):
        return (self.bbox[0] + self.bbox[1]) / 2

    @property
    def bbox(self):
        res = np.concatenate([np.array(x.bbox) for x in self.id_map.values()])
        return res.min(0), res.max(0)

    @property
    def num_polygons(self):
        return sum([x.num_polygons for x in self.id_map.values()])

    def to_custom_json(self):
        return {
            "name": self.name,
            "rawdata": self.rawdata,
            "guiParam": self.guiparam,
            "id_map": {key: value.to_custom_json() for key, value in self.id_map.items()},
        }

    def merge_by_color(self):
        from ada.ifc.utils import create_guid

        from .utils import merge_mesh_objects, organize_by_colour

        colour_map = organize_by_colour(self.id_map.values())

        id_map = dict()
        for colour, elements in colour_map.items():
            guid = create_guid()
            pm = merge_mesh_objects(elements)
            if len(pm.faces) == 0:
                continue
            id_map[guid] = pm

        return PartMesh(name=self.name, id_map=id_map)

    def __add__(self, other: PartMesh):
        self.id_map.update(other.id_map)
        return self


@dataclass
class ObjectMesh:
    guid: str
    faces: np.ndarray
    position: np.ndarray
    normal: np.ndarray | None = None
    color: list | None = None
    edges: np.ndarray = None
    vertex_color: np.ndarray = None
    instances: np.ndarray | None = None
    id_sequence: dict = field(default_factory=dict)
    translation: np.ndarray = None

    def translate(self, translation):
        self.position += translation

    @property
    def num_polygons(self):
        return int(len(self.faces) / 3)

    @property
    def bbox(self):
        return self.position.min(0), self.position.max(0)

    def to_binary_json(self, dest_dir, skip_normals=False):
        from ada.ifc.utils import create_guid

        dest_dir = pathlib.Path(dest_dir).resolve().absolute()
        pos_guid = create_guid()
        norm_guid = create_guid()
        index_guid = create_guid()
        vertex_guid = create_guid() if self.vertex_color is not None else None
        os.makedirs(dest_dir, exist_ok=True)

        np.save(str(dest_dir / pos_guid), self.position_flat)
        if skip_normals is False:
            np.save(str(dest_dir / norm_guid), self.normal_flat)
        np.save(str(dest_dir / index_guid), self.index_flat)

        if vertex_guid is not None:
            np.save(str(dest_dir / vertex_guid), self.vertex_color.astype(dtype="float32").flatten())

        return dict(
            index=index_guid,
            position=pos_guid,
            normal=norm_guid if skip_normals is False else None,
            color=self.color,
            vertexColor=vertex_guid if vertex_guid is not None else None,
            instances=self.instances,
            id_sequence=self.id_sequence,
            translation=self.translation_norm,
        )

    def to_custom_json(self):
        return dict(
            index=self.index_norm_flat,
            position=self.position_norm_flat,
            normal=self.normal_norm_flat,
            color=self.color,
            vertexColor=self.vertex_color_norm,
            instances=self.instances,
            id_sequence=self.id_sequence,
            translation=self.translation_norm,
        )

    def to_trimesh(self) -> list[trimesh.Trimesh]:
        from trimesh.visual.material import PBRMaterial

        indices_shape = get_shape(self.faces)
        verts_shape = get_shape(self.position)

        if indices_shape == 1:
            faces = self.faces.reshape(int(len(self.faces) / 3), 3)
        else:
            faces = self.faces

        if verts_shape == 1:
            vertices = self.position.reshape(int(len(self.position) / 3), 3)
        else:
            vertices = self.position

        vertex_color = None
        if self.vertex_color is not None:
            verts_shape = get_shape(self.vertex_color)
            if verts_shape == 1:
                vcolor = self.vertex_color.reshape(int(len(self.vertex_color) / 3), 3)
            else:
                vcolor = self.vertex_color

            vertex_color = np.array([[i * 255 for i in x] + [1] for x in vcolor], dtype=np.uint8)
            # vertex_color = [int(x * 255) for x in self.vertex_color]

        # vertex_normals = self.normal
        new_mesh = trimesh.Trimesh(
            vertices=vertices,
            faces=faces,
            # vertex_normals=vertex_normals,
            metadata=dict(guid=self.guid),
            vertex_colors=vertex_color,
        )
        if vertex_color is not None:
            new_mesh.visual.material = PBRMaterial(doubleSided=True)

        if self.color is not None:
            needs_to_be_scaled = True
            for x in self.color:
                if x > 1.0:
                    needs_to_be_scaled = False

            if needs_to_be_scaled:
                base_color = [int(x * 255) for x in self.color[:3]] + [int(self.color[3])]
            else:
                base_color = self.color

            if vertex_color is None:
                new_mesh.visual.material = PBRMaterial(baseColorFactor=base_color)

        meshes = [new_mesh]
        if self.edges is not None:
            from trimesh.path.entities import Line

            shape_edges = get_shape(self.edges)
            if shape_edges == 1:
                reshaped = self.edges.reshape(int(len(self.edges) / 2), 2)
            elif shape_edges == 2:
                reshaped = self.edges
            else:
                raise NotImplementedError("Edges consisting of more than 2 vertices is not supported")

            entities = [Line(x) for x in reshaped]
            edge_mesh = trimesh.path.Path3D(entities=entities, vertices=vertices)
            meshes.append(edge_mesh)

        return meshes

    @property
    def index_flat(self):
        return self.faces.astype(dtype="int32").flatten()

    @property
    def index_norm_flat(self):
        return self.index_flat.astype(dtype="int32").tolist()

    @property
    def position_flat(self):
        return self.position.astype(dtype="float32").flatten()

    @property
    def position_norm_flat(self):
        return self.position_flat.tolist()

    @property
    def normal_flat(self):
        return self.normal.astype(dtype="float32").flatten() if self.normal is not None else self.normal

    @property
    def normal_norm_flat(self):
        return self.normal_flat.tolist() if self.normal is not None else self.normal

    @property
    def vertex_color_norm(self):
        return self.vertex_color.astype(dtype="float32").tolist() if self.vertex_color is not None else None

    @property
    def translation_norm(self):
        return self.translation.astype(dtype="float32").tolist() if self.translation is not None else None

    def __add__(self, other: ObjectMesh):
        pos_len = int(len(self.position))
        new_index = other.faces + pos_len
        ma = int((len(other.faces) + len(self.faces))) - 1
        mi = int(len(self.faces))

        if len(self.faces) == 0:
            self.faces = new_index
        else:
            self.faces = np.concatenate([self.faces, new_index])

        if len(self.position) == 0:
            self.position = other.position
        else:
            self.position = np.concatenate([self.position, other.position])

        if self.color is None:
            self.color = other.color
        else:
            if other.color[-1] == 1.0 and self.color[-1] != 1.0:
                logger.warning("Will merge colors with different opacity.")
                self.color[-1] = 1.0

        if self.translation is None and other.translation is not None:
            self.translation = other.translation

        if self.normal is None or other.normal is None:
            self.normal = None
        else:
            if len(self.normal) == 0:
                self.normal = other.normal
            else:
                self.normal = np.concatenate([self.normal, other.normal])

        self.id_sequence[other.guid] = (mi, ma)
        return self


@dataclass
class VisNode:
    guid: str
    parent: str


def get_shape(np_array: np.ndarray) -> int:
    if len(np_array.shape) == 1:
        shape = len(np_array.shape)
    else:
        shape = np_array.shape[1]
    return shape



from collections.abc import Iterable
from typing import TYPE_CHECKING

from .helper_utils import get_instance_name
from .write_orientations import csys_str

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem import Connector, ConnectorSection


def connectors_str(fem: "FEM") -> str:
    return "\n".join([connector_str(con, True) for con in fem.elements.connectors])


def connector_sections_str(fem: "FEM") -> str:
    return "\n".join([connector_section_str(consec) for consec in fem.connector_sections.values()])


def connector_str(connector: "Connector", written_on_assembly_level: bool) -> str:
    csys_ref = "" if connector.csys is None else f'\n "{connector.csys.name}",'

    end1 = get_instance_name(connector.n1, written_on_assembly_level)
    end2 = get_instance_name(connector.n2, written_on_assembly_level)
    return f"""**
** ----------------------------------------------------------------
** Connector element representing {connector.name}
** ----------------------------------------------------------------
**
*Elset, elset={connector.name}
 {connector.id},
*Element, type=CONN3D2
 {connector.id}, {end1}, {end2}
*Connector Section, elset={connector.name}, behavior={connector.con_sec.name}
 {connector.con_type},{csys_ref}
**
{csys_str(connector.csys, written_on_assembly_level)}
**"""


def connector_section_str(con_sec: "ConnectorSection") -> str:
    conn_txt = """*Connector Behavior, name={0}""".format(con_sec.name)
    elast = con_sec.elastic_comp
    damping = con_sec.damping_comp
    plastic_comp = con_sec.plastic_comp
    rigid_dofs = con_sec.rigid_dofs
    soft_elastic_dofs = con_sec.soft_elastic_dofs
    if type(elast) is float:
        conn_txt += """\n*Connector Elasticity, component=1\n{0:.3E},""".format(elast)
    else:
        for i, comp in enumerate(elast):
            if isinstance(comp, Iterable) is False:
                conn_txt += """\n*Connector Elasticity, component={1} \n{0:.3E},""".format(comp, i + 1)
            else:
                conn_txt += f"\n*Connector Elasticity, nonlinear, component={i + 1}, DEPENDENCIES=1"
                for val in comp:
                    conn_txt += "\n" + ", ".join([f"{x:>12.3E}" if u <= 1 else f",{x:>12d}" for u, x in enumerate(val)])

    if type(damping) is float:
        conn_txt += """\n*Connector Damping, component=1\n{0:.3E},""".format(damping)
    else:
        for i, comp in enumerate(damping):
            if type(comp) is float:
                conn_txt += """\n*Connector Damping, component={1} \n{0:.3E},""".format(comp, i + 1)
            else:
                conn_txt += """\n*Connector Damping, nonlinear, component=1, DEPENDENCIES=1"""
                for val in comp:
                    conn_txt += "\n" + ", ".join(
                        ["{:>12.3E}".format(x) if u <= 1 else ",{:>12d}".format(x) for u, x in enumerate(val)]
                    )

    # Optional Choices
    if plastic_comp is not None:
        for i, comp in enumerate(plastic_comp):
            conn_txt += """\n*Connector Plasticity, component={}\n*Connector Hardening, definition=TABULAR""".format(
                i + 1
            )
            for val in comp:
                force, motion, rate = val
                conn_txt += "\n{}, {}, {}".format(force, motion, rate)

    if rigid_dofs is not None:
        conn_txt += "\n*Connector Elasticity, rigid\n "
        conn_txt += ", ".join(["{0}".format(x) for x in rigid_dofs])

    if soft_elastic_dofs is not None:
        for dof in soft_elastic_dofs:
            conn_txt += "\n*Connector Elasticity, component={0}\n 5.0,\n".format(dof)

    return conn_txt



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.sat.exceptions import InsufficientPointsError

if TYPE_CHECKING:
    from ada.sat.factory import SatStore


class PlateFactory:
    # Face row
    name_idx = 2
    loop_idx = 7

    # Loop row
    coedge_ref = 7

    def __init__(self, sat_store: SatStore):
        self.sat_store = sat_store

    def get_face_name_and_points(self, face_data_str: str) -> tuple[str, list[tuple[float]]]:
        res = face_data_str.strip().split()

        name = self.sat_store.get_name(res[self.name_idx])
        if not name.startswith("FACE"):
            raise NotImplementedError(f"Only face_refs starting with 'FACE' is supported. Found {name}")

        edges = self.get_edges(res)
        points = self.get_points(edges)
        return name, points

    def get_points(self, edges: list[list[str]]) -> list[tuple[float]]:
        p1, p2 = self.get_points_from_edge(edges[0])

        points = [p1, p2]

        for coedge in edges:
            p1, p2 = self.get_points_from_edge(coedge)
            edge_direction = str(coedge[-4])
            if edge_direction == "forward":
                p = p2
            else:
                p = p1
            if p not in points:
                points.append(p)

        if len(points) < 3:
            raise InsufficientPointsError("Plates cannot have < 3 points")

        coedge_first_direction = str(edges[0][-4])
        if coedge_first_direction == "reversed":
            points.reverse()

        return points

    def get_edges(self, face_data_list: list[str]) -> list[list[str]]:
        loop = self.sat_store.get(face_data_list[self.loop_idx])
        coedge_start_id = loop[self.coedge_ref]
        coedge_first = self.sat_store.get(coedge_start_id)

        coedge_first_direction = str(coedge_first[-4])

        # Coedge row
        next_coedge_idx = 6 if coedge_first_direction == "forward" else 7

        next_coedge = True
        coedge_next_id = coedge_first[next_coedge_idx]
        edges = [coedge_first]

        max_iter = 500
        i = 0
        while next_coedge is True:
            coedge = self.sat_store.get(coedge_next_id)
            edges.append(coedge)

            coedge_next_id = coedge[next_coedge_idx]
            if coedge_next_id == coedge_start_id:
                next_coedge = False

            i += 1
            if i > max_iter:
                raise ValueError(f"Found {i} points which is over max={max_iter}")
        return edges

    def get_points_from_edge(self, coedge: list[str]):
        # Coedge row
        edge_ref = 9

        # Edge row
        vert1_idx = 6
        vert2_idx = 8
        # edge_type_idx = 8

        # Vertex row
        p_idx = -2

        edge = self.sat_store.get(coedge[edge_ref])
        vert1 = self.sat_store.get(edge[vert1_idx])
        vert2 = self.sat_store.get(edge[vert2_idx])
        # edge_type = get_value_from_satd(edge[edge_type_idx], satd)
        p1 = self.sat_store.get(vert1[p_idx])
        p2 = self.sat_store.get(vert2[p_idx])
        n1 = tuple([float(x) for x in p1[-4:-1]])
        n2 = tuple([float(x) for x in p2[-4:-1]])
        return n1, n2



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ada.base.changes import ChangeAction
from ada.config import get_logger
from ada.ifc.read.read_physical_objects import import_physical_ifc_elem
from ada.ifc.read.reader_utils import (
    add_to_assembly,
    get_ifc_property_sets,
    get_parent,
    resolve_name,
)

from .read_materials import MaterialImporter
from .read_parts import PartImporter

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


@dataclass
class IfcReader:
    ifc_store: IfcStore

    def load_spatial_hierarchy(self):
        pi = PartImporter(self.ifc_store)
        pi.load_hierarchies()

    def load_materials(self):
        mi = MaterialImporter(self.ifc_store)
        mi.load_ifc_materials()

    def load_presentation_layers(self):
        from ada.concepts.presentation_layers import (
            PresentationLayer,
            PresentationLayers,
        )

        layers = dict()
        for obj in self.ifc_store.f.by_type("IfcPresentationLayerAssignment"):
            members = []
            for x in obj.AssignedItems:
                guid = None
                for inverse in self.ifc_store.f.get_inverse(x):
                    if inverse.is_a("IfcShapeRepresentation"):
                        product = inverse.OfProductRepresentation[0].ShapeOfProduct[0]
                        guid = product.GlobalId
                        break
                elem = self.ifc_store.assembly.get_by_guid(guid)
                if elem is None:
                    # raise ValueError()
                    continue
                if elem not in members:
                    members.append(elem)

            pl = PresentationLayer(
                obj.Name, obj.Description, members, identifier=obj.Identifier, change_type=ChangeAction.NOCHANGE
            )
            layers[pl.name] = pl

        self.ifc_store.assembly.presentation_layers = PresentationLayers(layers)

    def load_objects(self, data_only=False, elements2part=None):
        for product in self.ifc_store.f.by_type("IfcProduct"):
            if product.Representation is None or data_only is True:
                logger.info(f'Passing product "{product}"')
                continue

            parent = get_parent(product)
            name = product.Name

            props = get_ifc_property_sets(product)

            if name is None:
                name = resolve_name(props, product)

            logger.info(f"importing {name}")

            obj = import_physical_ifc_elem(product, name, self.ifc_store)
            if obj is None:
                continue

            obj.metadata = props
            add_to_assembly(self.ifc_store.assembly, obj, parent, elements2part)



from __future__ import annotations

import os
import pathlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Iterable

import meshio
import numpy as np

from ada.config import get_logger
from ada.fem.formats.general import FEATypes
from ada.fem.shapes.definitions import LineShapes, MassTypes, ShellShapes, SolidShapes

from .field_data import ElementFieldData, NodalFieldData, NodalFieldType

if TYPE_CHECKING:
    from ada import Material, Node, Section
    from ada.fem import Elem, FemSet
    from ada.fem.results.concepts import EigenDataSummary

logger = get_logger()


@dataclass
class ElementInfo:
    type: LineShapes | SolidShapes | ShellShapes
    source_software: FEATypes
    source_type: str | int


@dataclass
class ElementBlock:
    elem_info: ElementInfo
    node_refs: np.ndarray
    identifiers: np.ndarray


@dataclass
class FemNodes:
    coords: np.ndarray
    identifiers: np.ndarray

    def get_node_by_id(self, node_id: int | list[int]) -> list[Node]:
        from typing import Iterable

        from ada import Node

        if isinstance(node_id, Iterable) is False:
            node_id = [node_id]

        node_indices = [np.where(self.identifiers == x)[0][0] for x in node_id]
        return [Node(x, int(node_id[i])) for i, x in enumerate(self.coords[node_indices])]


@dataclass
class Mesh:
    elements: list[ElementBlock]
    nodes: FemNodes

    sections: dict[int, Section] = None
    materials: dict[int, Material] = None
    vectors: dict[int, list] = None
    elem_data: np.ndarray = None  # el_id, mat_id, sec_id, vec_id
    sets: dict[str, FemSet] = None

    def get_elem_by_id(self, elem_id: int) -> Elem:
        from ada.base.types import GeomRepr
        from ada.fem import Elem, FemSection, FemSet

        el_id, mat_id, sec_id, vec_id = self.elem_data[np.where(self.elem_data[:, 0] == elem_id)[0], :][0]
        mat = self.materials.get(int(mat_id))
        sec = self.sections.get(int(sec_id))
        vec = self.vectors.get(int(vec_id))

        elem = None
        for block in self.elements:
            res = np.where(block.identifiers == elem_id)
            for node_ids in block.node_refs[res]:
                nodes = self.nodes.get_node_by_id(node_ids)
                elem = Elem(elem_id, nodes, block.elem_info.type)
                break

        fs = FemSection(f"FS{sec_id}", GeomRepr.LINE, FemSet(f"El{el_id}", [elem]), mat, sec, local_z=vec)
        elem.fem_sec = fs
        return elem

    def get_edges_and_faces_from_mesh(self) -> tuple[np.ndarray, np.ndarray]:
        from ada.fem.shapes import ElemShape
        from ada.fem.shapes import definitions as shape_def

        nmap = {x: i for i, x in enumerate(self.nodes.identifiers)}
        keys = np.array(list(nmap.keys()))

        edges = []
        faces = []
        for cell_block in self.elements:
            el_type = cell_block.elem_info.type

            nodes_copy = cell_block.node_refs.copy()
            nodes_copy[np.isin(nodes_copy, keys)] = np.vectorize(nmap.get)(nodes_copy[np.isin(nodes_copy, keys)])

            for elem in nodes_copy:
                elem_shape = ElemShape(el_type, elem)
                if elem_shape.type in (MassTypes.MASS,):
                    continue
                try:
                    edges += elem_shape.edges
                except IndexError as e:
                    logger.error(e)
                    continue
                if isinstance(elem_shape.type, shape_def.LineShapes):
                    continue
                faces += elem_shape.get_faces()

        faces = np.array(faces).reshape(int(len(faces) / 3), 3)
        edges = np.array(edges).reshape(int(len(edges) / 2), 2)
        return edges, faces


@dataclass
class FEAResult:
    name: str
    software: str | FEATypes
    results: list[ElementFieldData | NodalFieldData]
    mesh: Mesh
    results_file_path: pathlib.Path = None
    step_name_map: dict[int | float, str] = None
    description: str = None

    def __post_init__(self):
        if self.results is None:
            self.results = []

        for res in self.results:
            res._mesh = self.mesh

    def get_steps(self):
        steps = []
        for x in self.results:
            if x.step not in steps:
                steps.append(x.step)
        return steps

    def get_results_grouped_by_field_value(self) -> dict[str, list[ElementFieldData | NodalFieldData]]:
        results = dict()
        for x in self.results:
            if x.name not in results.keys():
                results[x.name] = []
            results[x.name].append(x)
        return results

    def get_data_by_field_and_elem_ids(
        self, field: str, elem_ids: list[int], int_points: list[int] = None
    ) -> list[ElementFieldData]:
        data = self.get_results_grouped_by_field_value()
        values = data.get(field)
        output_res = []
        for sdata in values:
            output_res.append(sdata.get_by_element_id(elem_ids, int_points))

        return output_res

    def get_data_by_field_name_and_set_name(self, field, set_name, int_points=None) -> list[ElementFieldData]:
        fs = self.mesh.sets.get(set_name)
        return self.get_data_by_field_and_elem_ids(field, fs.members, int_points)

    def get_field_value_by_name(
        self, name: str, step: int = None
    ) -> ElementFieldData | NodalFieldData | list[ElementFieldData | NodalFieldData]:
        data = self.get_results_grouped_by_field_value()
        values = data.get(name)
        if values is None:
            raise ValueError(f"Unable to find field data '{name}'. Available are {list(data.keys())}")

        if len(values) == 1:
            return values[0]

        if step is not None:
            val_selected = [val for val in values if val.step == step]
            if val_selected == 0:
                raise ValueError(f"Unable to find step=={step}. Available steps are {[x.step for x in values]}")

            return val_selected[0]

        return values

    def iter_results_by_field_value(self) -> Iterable[ElementFieldData | NodalFieldData]:
        for x in self.results:
            yield x

    def get_data(self, field: str, step: int):
        steps = self.get_results_grouped_by_field_value().get(field)
        if step == -1:
            field_data = list(sorted(steps, key=lambda x: x.step))[-1]
        else:
            all_field_data = [x for x in steps if x.step == step]
            if len(all_field_data) != 1:
                raise ValueError(
                    f"Found {len(all_field_data)} results of field data based on step {step}.\n"
                    f"Available steps are {[x.step for x in steps]}"
                )

            field_data = all_field_data[0]

        return field_data.get_all_values()

    def _get_cell_blocks(self):
        cells = []
        for cb in self.mesh.elements:
            cell_type = cb.elem_info.type.value.lower()
            ncopy = cb.node_refs.copy()
            for i, v in enumerate(self.mesh.nodes.identifiers):
                ncopy[np.where(ncopy == v)] = i
            cells += [meshio.CellBlock(cell_type=cell_type, data=ncopy)]
        return cells

    def _get_point_and_cell_data(self) -> tuple[dict, dict]:
        from .field_data import ElementFieldData, NodalFieldData

        cell_data = dict()
        point_data = dict()
        for values in self.get_results_grouped_by_field_value().values():
            for x in values:
                res = x.get_all_values()
                name = f"{x.name} - {x.step}" if len(values) > 1 else x.name
                if isinstance(x, NodalFieldData):
                    point_data[name] = res
                elif isinstance(x, ElementFieldData) and x.field_pos == x.field_pos.NODAL:
                    point_data[name] = res
                elif isinstance(x, ElementFieldData) and x.field_pos == x.field_pos.INT:
                    if isinstance(res, dict):
                        cell_data.update(res)
                    else:
                        cell_data[name] = [res]
                else:
                    raise ValueError()

        return cell_data, point_data

    def _colorize_data(self, field: str, step: int, colorize_function: Callable = None):
        from ada.visualize.colors import DataColorizer

        data = self.get_data(field, step)
        vertex_colors = DataColorizer.colorize_data(data, func=colorize_function)
        return np.array([[i * 255 for i in x] + [1] for x in vertex_colors], dtype=np.int32)

    def _warp_data(self, vertices: np.ndarray, field: str, step, scale: float = 1.0):
        data = self.get_data(field, step)

        result = vertices + data[:, :3] * scale
        return result

    def to_meshio_mesh(self) -> meshio.Mesh:
        cells = self._get_cell_blocks()
        cell_data, point_data = self._get_point_and_cell_data()

        return meshio.Mesh(points=self.mesh.nodes.coords, cells=cells, cell_data=cell_data, point_data=point_data)

    def to_xdmf(self, filepath):
        cells = self._get_cell_blocks()
        with meshio.xdmf.TimeSeriesWriter(filepath) as writer:
            writer.write_points_cells(self.mesh.nodes.coords, cells)
            for key, values in self.get_results_grouped_by_field_value().items():
                for x in values:
                    res = x.get_all_values()
                    name = x.name
                    point_data = dict()
                    if isinstance(x, NodalFieldData):
                        point_data[name] = res
                    elif isinstance(x, ElementFieldData) and x.field_pos == x.field_pos.NODAL:
                        point_data[name] = res
                    elif isinstance(x, ElementFieldData) and x.field_pos == x.field_pos.INT:
                        raise NotImplementedError("Currently not supporting element data directly from int. points")
                    else:
                        raise ValueError()

                    writer.write_data(x.step, point_data=point_data)

    def to_fem_file(self, fem_file: str | pathlib.Path):
        if isinstance(fem_file, str):
            fem_file = pathlib.Path(fem_file)

        mesh = self.to_meshio_mesh()
        mesh.write(fem_file)

    def to_trimesh(self, step: int, field: str, warp_field=None, warp_step=None, warp_scale=None, cfunc=None):
        import trimesh
        from trimesh.path.entities import Line
        from trimesh.visual.material import PBRMaterial

        vertices = self.mesh.nodes.coords
        edges, faces = self.mesh.get_edges_and_faces_from_mesh()

        # Colorize data
        vertex_color = self._colorize_data(field, step, cfunc)

        # Warp data
        if warp_field is not None:
            warped_vertices = self._warp_data(vertices, warp_field, warp_step, warp_scale)
            vertices = warped_vertices

        new_mesh = trimesh.Trimesh(vertices=vertices, faces=faces, vertex_colors=vertex_color)
        new_mesh.visual.material = PBRMaterial(doubleSided=True)

        entities = [Line(x) for x in edges]
        edge_mesh = trimesh.path.Path3D(entities=entities, vertices=vertices)

        scene = trimesh.Scene()
        scene.add_geometry(new_mesh, node_name=self.name, geom_name="faces")
        scene.add_geometry(edge_mesh, node_name=f"{self.name}_edges", geom_name="edges", parent_node_name=self.name)
        return scene

    def to_gltf(self, dest_file, step: int, field: str, warp_field=None, warp_step=None, warp_scale=None, cfunc=None):
        from ada.core.vector_utils import rot_matrix

        dest_file = pathlib.Path(dest_file).resolve().absolute()
        scene = self.to_trimesh(step, field, warp_field, warp_step, warp_scale, cfunc)

        # Trimesh automatically transforms by setting up = Y. This will counteract that transform
        m3x3 = rot_matrix((0, -1, 0))
        m3x3_with_col = np.append(m3x3, np.array([[0], [0], [0]]), axis=1)
        m4x4 = np.r_[m3x3_with_col, [np.array([0, 0, 0, 1])]]
        scene.apply_transform(m4x4)

        os.makedirs(dest_file.parent, exist_ok=True)
        print(f'Writing Visual Mesh to "{dest_file}"')
        with open(dest_file, "wb") as f:
            scene.export(file_obj=f, file_type=dest_file.suffix[1:])

    def get_eig_summary(self) -> EigenDataSummary:
        """If the results are eigenvalue results, this method will return a summary of the eigenvalues and modes"""
        from ada.fem.results.eigenvalue import EigenDataSummary, EigenMode

        modes = []
        for x in self.results:
            if isinstance(x, NodalFieldData) and x.field_type != NodalFieldType.DISP:
                continue
            m = EigenMode(x.step, f_hz=x.eigen_freq, eigenvalue=x.eigen_value)
            modes.append(m)
        return EigenDataSummary(modes)



from __future__ import annotations

import os
from io import StringIO
from typing import TYPE_CHECKING

from .write_amplitudes import amplitudes_str
from .write_bc import boundary_conditions_str
from .write_connectors import connector_sections_str, connectors_str
from .write_constraints import constraints_str
from .write_elements import elements_str
from .write_interactions import eval_interactions, int_prop_str
from .write_main_inp import write_main_inp_str
from .write_masses import masses_str
from .write_materials import materials_str
from .write_nodes import nodes_str
from .write_orientations import orientations_str
from .write_parts import write_all_parts
from .write_predefined_state import predefined_fields_str
from .write_sets import elsets_str, nsets_str
from .write_steps import write_step
from .write_surfaces import surfaces_str

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly

__all__ = ["to_fem"]


def to_fem(assembly: Assembly, name, analysis_dir=None, metadata=None, writable_obj: StringIO = None):
    """Build the Abaqus Analysis input deck"""

    # Write part bulk files
    write_all_parts(assembly, analysis_dir)

    # Write Assembly level files
    core_dir = analysis_dir / r"core_input_files"
    os.makedirs(core_dir)

    afem = assembly.fem

    # Main Input File
    with open(analysis_dir / f"{name}.inp", "w") as d:
        d.write(write_main_inp_str(assembly, analysis_dir))

    # Connector Sections
    with open(core_dir / "connector_sections.inp", "w") as d:
        d.write(connector_sections_str(afem))

    # Connectors
    with open(core_dir / "connectors.inp", "w") as d:
        d.write(connectors_str(afem) if len(list(afem.elements.connectors)) > 0 else "**")

    # Constraints
    with open(core_dir / "constraints.inp", "w") as d:
        d.write(constraints_str(afem, True) if len(afem.constraints.keys()) > 0 else "**")

    # Assembly data
    with open(core_dir / "assembly_data.inp", "w") as d:
        assembly_nodes_str = "** No Nodes"
        if len(afem.nodes) > 0:
            assembly_nodes_str = nodes_str(afem)
        d.write(f"{assembly_nodes_str}\n")
        d.write(f"{nsets_str(afem, True)}\n")
        d.write(f"{elsets_str(afem, True)}\n")
        d.write(f"{surfaces_str(afem, True)}\n")
        d.write(orientations_str(afem, True) + "\n")
        d.write(elements_str(afem, True) + "\n")
        d.write(masses_str(afem, True))

    # Amplitude data
    with open(core_dir / "amplitude_data.inp", "w") as d:
        d.write(amplitudes_str(afem))

    # Interaction Properties
    with open(core_dir / "interaction_prop.inp", "w") as d:
        d.write(int_prop_str(afem))

    # Interactions data
    eval_interactions(assembly, analysis_dir)
    with open(core_dir / "interactions.inp", "a") as d:
        d.write(predefined_fields_str(afem))

    # Materials data
    with open(core_dir / "materials.inp", "w") as d:
        d.write(materials_str(assembly))

    # Boundary Condition data
    with open(core_dir / "bc_data.inp", "w") as d:
        d.write(boundary_conditions_str(assembly))

    # Analysis steps
    for step_in in afem.steps:
        write_step(step_in, analysis_dir)

    print(f'Created an Abaqus input deck at "{analysis_dir}"')



from ada import Part, Section
from ada.concepts.containers import Sections
from ada.sections import GeneralProperties


def get_sections(xml_root, parent: Part) -> Sections:
    all_secs = xml_root.findall(".//section")
    sections = [interpret_section_props(sec_el.attrib["name"], sec_el[0], parent) for sec_el in all_secs]
    return Sections(sections, parent=parent)


def interpret_section_props(name, sec_prop, parent: Part) -> Section:
    sec_map = dict(
        box_section=box_sec,
        i_section=isec,
        l_section=angular,
        unsymmetrical_i_section=unsymm_isec,
        pipe_section=pipe_section,
        channel_section=channel_section,
        bar_section=bar_section,
        general_section=general_section,
        cone_section=cone_section,
        pgb_section=pgb_section,
    )
    sec_interpreter = sec_map.get(sec_prop.tag, None)

    if sec_interpreter is None:
        raise ValueError(f"Missing property {sec_prop.tag}")

    section = sec_interpreter(name, sec_prop)
    section.parent = parent

    return section


def box_sec(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.BOX,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_top=float(sec_prop.attrib["b"]),
        w_btn=float(sec_prop.attrib["b"]),
        t_w=float(sec_prop.attrib["tw"]),
        t_ftop=float(sec_prop.attrib["tftop"]),
        t_fbtn=float(sec_prop.attrib["tfbot"]),
    )


def angular(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.ANGULAR,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_btn=float(sec_prop.attrib["b"]),
        t_w=float(sec_prop.attrib["tw"]),
        t_fbtn=float(sec_prop.attrib["tf"]),
    )


def isec(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.IPROFILE,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_top=float(sec_prop.attrib["b"]),
        w_btn=float(sec_prop.attrib["b"]),
        t_w=float(sec_prop.attrib["tw"]),
        t_ftop=float(sec_prop.attrib["tf"]),
        t_fbtn=float(sec_prop.attrib["tf"]),
    )


def unsymm_isec(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.IPROFILE,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_btn=float(sec_prop.attrib["bfbot"]),
        w_top=float(sec_prop.attrib["bftop"]),
        t_w=float(sec_prop.attrib["tw"]),
        t_ftop=float(sec_prop.attrib["tftop"]),
        t_fbtn=float(sec_prop.attrib["tfbot"]),
    )


def pipe_section(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.TUBULAR,
        sec_str=name,
        r=float(sec_prop.attrib["od"]) / 2,
        wt=float(sec_prop.attrib["th"]),
    )


def channel_section(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.CHANNEL,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_top=float(sec_prop.attrib["b"]),
        w_btn=float(sec_prop.attrib["b"]),
        t_w=float(sec_prop.attrib["tw"]),
        t_ftop=float(sec_prop.attrib["tf"]),
        t_fbtn=float(sec_prop.attrib["tf"]),
    )


def bar_section(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_type=Section.TYPES.FLATBAR,
        sec_str=name,
        h=float(sec_prop.attrib["h"]),
        w_top=float(sec_prop.attrib["b"]),
        w_btn=float(sec_prop.attrib["b"]),
    )


def general_section(name, sec_prop) -> Section:
    return Section(
        name=name,
        sec_str=name,
        sec_type=Section.TYPES.GENERAL,
        genprops=GeneralProperties(
            Ax=float(sec_prop.attrib["area"]),
            Ix=float(sec_prop.attrib["ix"]),
            Iy=float(sec_prop.attrib["iy"]),
            Iz=float(sec_prop.attrib["iz"]),
            Iyz=float(sec_prop.attrib["iyz"]),
            Wxmin=float(sec_prop.attrib["wxmin"]),
            Wymin=float(sec_prop.attrib["wymin"]),
            Wzmin=float(sec_prop.attrib["wzmin"]),
            Shary=float(sec_prop.attrib["shary"]),
            Sharz=float(sec_prop.attrib["sharz"]),
            Shceny=float(sec_prop.attrib["shceny"]),
            Shcenz=float(sec_prop.attrib["shcenz"]),
            Sy=float(sec_prop.attrib["sy"]),
            Sz=float(sec_prop.attrib["sz"]),
            Sfy=float(sec_prop.attrib["sfy"]),
            Sfz=float(sec_prop.attrib["sfz"]),
        ),
    )


def cone_section(name, sec_prop) -> Section:
    return Section(name, sec_type=Section.TYPES.GENERAL, genprops=GeneralProperties(Ax=0.1))


def pgb_section(name, sec_prop):
    # circ = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
    h, b, tw, otw, tf = [float(sec_prop.attrib[x]) for x in ("h", "b", "tw", "otw", "tf")]

    # ot = [(x * h / 2, y * b / 2) for x, y in circ]
    # it1 = [(x * h / 2 + otw, y * b / 2 + tf) for x, y in circ]
    # it2 = [(b / 2 + tw + x * h / 2 + otw, y * b / 2 + tf) for x, y in circ]

    return Section(name, Section.TYPES.BOX, h=h, w_btn=b, w_top=b, t_w=otw, t_fbtn=tf, t_ftop=tf)



from __future__ import annotations

from typing import TYPE_CHECKING

from ada import ArcSegment, Node, PipeSegElbow, PipeSegStraight
from ada.config import get_logger

from .read_beam_section import import_section_from_ifc
from .read_materials import read_material
from .reader_utils import (
    get_associated_material,
    get_axis_polyline_points_from_product,
    get_ifc_property_sets,
    get_swept_area,
)

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def import_pipe_segment(segment, name, ifc_store: IfcStore) -> PipeSegStraight | PipeSegElbow:
    if segment.is_a("IfcPipeSegment"):
        pipe_segment = read_pipe_straight_segment(segment, name, ifc_store)
    else:
        pipe_segment = read_pipe_elbow(segment, name, ifc_store)

    pipe_segment.section.parent = pipe_segment
    pipe_segment.material.parent = pipe_segment

    return pipe_segment


def read_pipe_straight_segment(segment, name, ifc_store: IfcStore) -> PipeSegStraight:
    p1, p2 = get_axis_polyline_points_from_product(segment)
    mat_ref = get_associated_material(segment)
    swept_area = get_swept_area(segment)
    section = import_section_from_ifc(swept_area)
    mat = read_material(mat_ref, ifc_store)

    pipe_segment = PipeSegStraight(name, Node(p1), Node(p2), section, mat, guid=segment.GlobalId)

    return pipe_segment


def read_pipe_elbow(segment, name, ifc_store: IfcStore) -> PipeSegElbow:
    p1, p2, p3 = [Node(x) for x in get_axis_polyline_points_from_product(segment)]
    pset = get_ifc_property_sets(segment)
    bend_radius = pset.get("Properties", dict()).get("bend_radius", None)
    arc_p1 = pset.get("Properties", dict()).get("p1", None)
    arc_p2 = pset.get("Properties", dict()).get("p2", None)
    arc_midpoint = pset.get("Properties", dict()).get("midpoint", None)

    if bend_radius is None or arc_midpoint is None:
        logger.error("The current Elbow interpretation requires a specific property with bend radius to be imported")
        return None

    bend_radius = float(bend_radius)
    mat_ref = get_associated_material(segment)
    swept_area = get_swept_area(segment)
    section = import_section_from_ifc(swept_area)
    mat = read_material(mat_ref, ifc_store)

    arc = ArcSegment(arc_p1, arc_p2, arc_midpoint, bend_radius)
    elbow = PipeSegElbow(name, p1, p2, p3, bend_radius, section, mat, arc_seg=arc, guid=segment.GlobalId)
    return elbow



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem import StepImplicit

from .static_lin import step_static_lin_str
from .static_nonlin import step_static_nonlin_str

if TYPE_CHECKING:
    from ada.concepts.spatial import Part


def step_static_str(step: StepImplicit, part: Part) -> str:
    if step.nl_geom is True:
        return step_static_nonlin_str(step, part)
    else:
        return step_static_lin_str(step, part)



def create_vertex_string(vertex_id, vertex):
    vertex_string = f"{vertex_id} vertex {vertex[0]} {vertex[1]} {vertex[2]} #\n"
    return vertex_string


def create_edge_string(edge_id, vertex_start_id, vertex_end_id, vertex):
    edge_string = (
        f"{edge_id} edge -1 -1 -1 -1 {vertex_start_id} 0 {vertex_end_id} 0.51249999999999996 "
        f"{vertex_start_id} {vertex_end_id} forward @7 unknown T"
        f" {vertex[0]} {vertex[1]} {vertex[2]} {vertex[0]} {vertex[1]} {vertex[2]} #\n"
    )
    return edge_string


def create_coedge_string(coedge_id, edge_id, loop_id, orientation):
    coedge_string = f"{coedge_id} coedge -1 -1 -1 -1 {edge_id} {edge_id} -1 {loop_id} {orientation} #\n"
    return coedge_string


def create_loop_string(loop_id, coedge_id):
    loop_string = (
        f"{loop_id} loop -1 -1 -1 -1 -1 {coedge_id} -1 T -0.375 "
        f"-0.28749999999999998 0 0.4375 0.22500000000000001 0 periphery #\n"
    )
    return loop_string


def create_face_string(face_id, loop_id):
    face_string = (
        f"{face_id} face -1 -1 -1 -1 -1 {loop_id} -1 -1 forward double out T "
        f"-0.375 -0.28749999999999998 0 0.4375 0.22500000000000001 0 F #\n"
    )
    return face_string


def create_shell_string(shell_id, face_id):
    shell_string = (
        f"{shell_id} shell -1 -1 -1 -1 -1 -1 {face_id} -1 T "
        f"-0.375 -0.28749999999999998 0 0.4375 0.22500000000000001 0 #\n"
    )
    return shell_string


def create_lump_string(lump_id, shell_id):
    lump_string = (
        f"{lump_id} lump -1 -1 -1 -1 -1 {shell_id} -1 T "
        f"-0.375 -0.28749999999999998 0 0.4375 0.22500000000000001 0 #\n"
    )
    return lump_string


def create_body_string(body_id, lump_id):
    body_string = (
        f"{body_id} body -1 {lump_id} -1 -1 -1 -1 T " f"-0.375 -0.28749999999999998 0 0.4375 0.22500000000000001 0 #\n"
    )
    return body_string



import xml.etree.ElementTree as ET

from ada import Node, Part
from ada.fem import Bc, FemSet


def get_boundary_conditions(xml_root: ET.Element, parent: Part) -> list[Bc]:
    bcs = []
    dof_map = dict(dx=1, dy=2, dz=3, rx=4, ry=5, rz=6)
    for sp in xml_root.findall(".//support_point"):
        name = sp.attrib.get("name")
        position = sp.findall(".//position")
        if len(position) != 1:
            raise NotImplementedError()

        # Get position
        pos = position[0]
        n = Node([float(y) for x, y in pos.items()])
        parent.fem.nodes.add(n)

        # get dofs
        dofs = []
        for dof in sp.findall(".//boundary_condition"):
            if dof.attrib["constraint"] == "fixed":
                dofs.append(dof_map.get(dof.attrib["dof"]))

        fs = FemSet(f"{name}_fs", [n])
        bc = Bc(name, fem_set=fs, dofs=dofs)
        bcs.append(bc)

    return bcs



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import meshio

from ada.fem import StepEigen

from .read_eigen_data import get_eigen_data

if TYPE_CHECKING:
    from ada.fem.results import Results

USE_FRDCONVERT = True


def read_calculix_results(results: Results, file_ref: pathlib.Path, overwrite) -> meshio.Mesh:
    from .read_frd_file import read_from_frd_file

    mesh = read_from_frd_file(file_ref)

    dat_file = file_ref.with_suffix(".dat")
    if results.assembly is not None and dat_file.exists() and isinstance(results.assembly.fem.steps[0], StepEigen):
        results.eigen_mode_data = get_eigen_data(dat_file)

    return mesh



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ada.base.root import Root
from ada.base.units import Units
from ada.config import Settings, get_logger
from ada.sections.categories import BaseTypes, SectionCat

if TYPE_CHECKING:
    from ada import Beam, CurvePoly, Pipe, PipeSegElbow, PipeSegStraight
    from ada.fem import FemSection

logger = get_logger()


class Section(Root):
    TYPES = BaseTypes

    def __init__(
        self,
        name,
        sec_type: BaseTypes | str = None,
        h=None,
        w_top=None,
        w_btn=None,
        t_w=None,
        t_ftop=None,
        t_fbtn=None,
        r=None,
        wt=None,
        sec_id=None,
        parent=None,
        sec_str=None,
        from_str=None,
        outer_poly: CurvePoly = None,
        inner_poly: CurvePoly = None,
        genprops: GeneralProperties = None,
        metadata=None,
        units=Units.M,
        guid=None,
        refs=None,
    ):
        super(Section, self).__init__(name=name, guid=guid, metadata=metadata, units=units, parent=parent)
        if isinstance(sec_type, str):
            sec_type_in = sec_type
            sec_type = BaseTypes.from_str(sec_type_in)
            if sec_type is None:
                raise ValueError(f"Unrecognized section type {sec_type_in}]")

        self._type = sec_type
        self._h = h
        self._w_top = w_top
        self._w_btn = w_btn
        self._t_w = t_w
        self._t_ftop = t_ftop
        self._t_fbtn = t_fbtn
        self._r = r
        self._wt = wt
        self._id = sec_id
        self._outer_poly = outer_poly
        self._inner_poly = inner_poly
        self._sec_str = sec_str

        self._ifc_profile = None
        self._ifc_beam_type = None

        if from_str is not None:
            from ada.sections.utils import interpret_section_str

            if units == Units.M:
                scalef = 0.001
            elif units == Units.MM:
                scalef = 1.0
            else:
                raise ValueError(f'Unknown units "{units}"')

            sec, tap = interpret_section_str(from_str, scalef, units=units)
            self.__dict__.update(sec.__dict__)

        self._genprops = None
        self._refs = refs if refs is not None else []
        if genprops is not None:
            genprops.parent = self
            self._genprops = genprops

    def equal_props(self, other: Section):
        for propa, propb in zip(self.unique_props(), other.unique_props()):
            if propa != propb:
                return False

        return True

    def unique_props(self):
        props = ["type", "h", "w_top", "w_btn", "t_w", "t_ftop", "t_fbtn", "r", "wt", "poly_outer", "poly_inner"]
        if self.type == self.TYPES.GENERAL:
            props += ["properties"]
        return tuple([getattr(self, p) for p in props])

    @property
    def type(self) -> BaseTypes:
        return self._type

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if type(value) is not int:
            raise ValueError
        self._id = value

    @property
    def h(self):
        return self._h

    @property
    def w_top(self):
        return self._w_top

    @w_top.setter
    def w_top(self, value):
        """Width of top flange"""
        self._w_top = value

    @property
    def w_btn(self):
        """Width of bottom flange"""
        return self._w_btn

    @w_btn.setter
    def w_btn(self, value):
        self._w_btn = value

    @property
    def t_w(self):
        """Thickness of web"""
        return self._t_w

    @property
    def t_ftop(self):
        """Thickness of top flange"""
        return self._t_ftop

    @property
    def t_fbtn(self):
        """Thickness of bottom flange"""
        return self._t_fbtn

    @property
    def r(self) -> float:
        """Radius (Outer)"""
        return self._r

    @r.setter
    def r(self, value: float):
        self._r = value
        self._genprops = None

    @property
    def wt(self) -> float:
        """Wall thickness"""
        return self._wt

    @wt.setter
    def wt(self, value: float):
        self._wt = value
        self._genprops = None

    @property
    def sec_str(self):
        def s(x):
            return x / 0.001

        if self.type == BaseTypes.BOX:
            sec_str = "{}{:g}x{:g}x{:g}x{:g}".format(
                self.type.value, s(self.h), s(self.w_top), s(self.t_w), s(self.t_ftop)
            )
        elif self.type == BaseTypes.TUBULAR:
            sec_str = "{}{:g}x{:g}".format(self.type.value, s(self.r), s(self.wt))
        elif self.type == BaseTypes.CIRCULAR:
            sec_str = "{}{:g}".format(self.type.value, s(self.r))
        elif self.type == BaseTypes.ANGULAR:
            sec_str = "{}{:g}x{:g}".format(self.type.value, s(self.h), s(self.t_w))
        elif self.type == BaseTypes.IPROFILE:
            sec_str = self._sec_str
        elif self.type == BaseTypes.TPROFILE:
            sec_str = "{}{:g}x{:g}x{:g}".format(self.type.value, s(self.h), s(self.w_top), s(self.t_w))
        elif self.type == BaseTypes.CHANNEL:
            sec_str = "{}{:g}".format(self.type.value, s(self.h))
        elif self.type == BaseTypes.GENERAL:
            sec_str = "{}{}".format(self.type.value, self.id)
        elif self.type == BaseTypes.FLATBAR:
            sec_str = f"{self.type.value}{s(self.h)}x{s(self.w_top)}"
        elif self.type == BaseTypes.POLY:
            sec_str = "PolyCurve"

        else:
            raise ValueError(f'Section type "{self.type}" has not been given a section str')

        return sec_str.replace(".", "_") if sec_str is not None else None

    @property
    def properties(self) -> GeneralProperties:
        if self._genprops is None:
            from .properties import calculate_general_properties

            self._genprops = calculate_general_properties(self)

        return self._genprops

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if self._units != value:
            scale_factor = Units.get_scale_factor(self._units, value)

            if self.poly_inner is not None:
                self.poly_inner.scale(scale_factor, Settings.point_tol)

            if self.poly_outer is not None:
                self.poly_outer.scale(scale_factor, Settings.point_tol)

            vals = ["h", "w_top", "w_btn", "t_w", "t_ftop", "t_fbtn", "r", "wt"]

            for key in self.__dict__.keys():
                if self.__dict__[key] is not None:
                    if key[1:] in vals:
                        self.__dict__[key] *= scale_factor
            self._units = value

    @property
    def poly_outer(self) -> CurvePoly:
        return self._outer_poly

    @property
    def poly_inner(self) -> CurvePoly:
        return self._inner_poly

    def get_section_profile(self, is_solid=True) -> SectionProfile:
        return build_section_profile(self, is_solid)

    def _repr_html_(self):
        from IPython.display import display
        from ipywidgets import HBox

        from ada.visualize.renderer_pythreejs import SectionRenderer

        sec_render = SectionRenderer()
        fig, html = sec_render.build_display(self)
        display(HBox([fig, html]))

    @property
    def refs(self) -> list[Beam | FemSection | Pipe | PipeSegStraight | PipeSegElbow]:
        return self._refs

    def __hash__(self):
        return hash(self.guid)

    def __repr__(self):
        if self.type in SectionCat.circular + SectionCat.tubular:
            return f"Section({self.name}, {self.type}, r: {self.r}, wt: {self.wt})"
        elif self.type in SectionCat.general:
            p = self.properties
            return f"Section({self.name}, {self.type}, Ax: {p.Ax}, Ix: {p.Ix}, Iy: {p.Iy}, Iz: {p.Iz}, Iyz: {p.Iyz})"
        else:
            return (
                f"Section({self.name}, {self.type}, h: {self.h}, w_btn: {self.w_btn}, "
                f"w_top: {self.w_top}, t_fbtn: {self.t_fbtn}, t_ftop: {self.t_ftop}, t_w: {self.t_w})"
            )


class SectionParts:
    WEB = "web"
    TOP_FLANGE = "top_fl"
    BTN_FLANGE = "btn_fl"


@dataclass
class GeneralProperties:
    parent: Section = field(default=None, compare=False, repr=False)
    Ax: float = None
    Ix: float = None
    Iy: float = None
    Iz: float = None
    Iyz: float = None
    Wxmin: float = None
    Wymin: float = None
    Wzmin: float = None
    Shary: float = None
    Sharz: float = None
    Shceny: float = None
    Shcenz: float = None
    Sy: float = None
    Sz: float = None
    Sfy: float = 1
    Sfz: float = 1
    Cy: float = None
    Cz: float = None

    def __hash__(self):
        return hash(
            (
                self.Ax,
                self.Ix,
                self.Iy,
                self.Iz,
                self.Iyz,
                self.Wxmin,
                self.Wymin,
                self.Wzmin,
                self.Shary,
                self.Sharz,
                self.Shceny,
                self.Shcenz,
                self.Sy,
                self.Sz,
                self.Sfy,
                self.Sfz,
                self.Cy,
                self.Cz,
            )
        )

    @property
    def modified(self) -> bool:
        """Returns true if attributes are not equal to the calculated properties of the parent section"""
        return self != self.calc_parent_properties()

    def calc_parent_properties(self) -> GeneralProperties:
        """Returns calculated properties based on parent section"""
        from ada.sections.properties import calculate_general_properties

        return calculate_general_properties(self.parent)


@dataclass
class SectionProfile:
    sec: Section
    is_solid: bool
    outer_curve: CurvePoly = None
    inner_curve: CurvePoly = None
    outer_curve_disconnected: list[CurvePoly] = None
    inner_curve_disconnected: list[CurvePoly] = None
    disconnected: bool = None
    shell_thickness_map: list[tuple[str, float]] = None


def build_section_profile(sec: Section, is_solid) -> SectionProfile:
    import ada.sections.profiles as profile_builder

    if sec.type in [BaseTypes.TUBULAR, BaseTypes.CIRCULAR, BaseTypes.GENERAL]:
        logger.info("Tubular profiles do not need curve representations")
        return SectionProfile(sec, is_solid)

    build_map = {
        BaseTypes.ANGULAR: profile_builder.angular,
        BaseTypes.IPROFILE: profile_builder.iprofiles,
        BaseTypes.TPROFILE: profile_builder.tprofiles,
        BaseTypes.BOX: profile_builder.box,
        BaseTypes.FLATBAR: profile_builder.flatbar,
        BaseTypes.CHANNEL: profile_builder.channel,
    }

    section_builder = build_map.get(sec.type, None)

    if section_builder is None and sec.poly_outer is None:
        raise ValueError("Currently geometry build is unsupported for profile type {ptype}".format(ptype=sec.type))

    if section_builder is not None:
        section_profile = section_builder(sec, is_solid)
    else:
        section_profile = SectionProfile(sec, outer_curve=sec.poly_outer, is_solid=is_solid, disconnected=False)

    return section_profile



from __future__ import annotations

from typing import TYPE_CHECKING

from ada import Placement, Plate
from ada.config import get_logger

from .read_curves import import_indexedpolycurve, import_polycurve
from .read_materials import read_material
from .reader_utils import get_associated_material

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def import_ifc_plate(ifc_elem, name, ifc_store: IfcStore) -> Plate:
    from .exceptions import NoIfcAxesAttachedError

    logger.info(f"importing {name}")
    ifc_mat = get_associated_material(ifc_elem)
    mat = None
    if ifc_store.assembly is not None:
        mat = ifc_store.assembly.get_by_name(name)

    if mat is None:
        mat = read_material(ifc_mat, ifc_store)

    # TODO: Fix interpretation of IfcIndexedPolyCurve. Should pass origin to get actual 2d coordinates.
    # Adding Axis information
    axes = [rep for rep in ifc_elem.Representation.Representations if rep.RepresentationIdentifier == "Axis"]
    if len(axes) != 1:
        raise NoIfcAxesAttachedError("IfcPlate does not have an Axis representation Item")
    axis = axes[0]
    origin = axis.Items[0].Points[0].Coordinates

    # Adding Body
    bodies = [rep for rep in ifc_elem.Representation.Representations if rep.RepresentationIdentifier == "Body"]
    if len(bodies) != 1:
        raise NotImplementedError("Geometry with multiple bodies is not currently supported")
    if len(bodies[0].Items) != 1:
        raise NotImplementedError("Body with multiple Items is not currently supported")

    item = bodies[0].Items[0]
    t = item.Depth
    normal = item.ExtrudedDirection.DirectionRatios
    xdir = item.Position.RefDirection.DirectionRatios
    outer_curve = item.SweptArea.OuterCurve

    if outer_curve.is_a("IfcIndexedPolyCurve"):
        nodes2d = import_indexedpolycurve(outer_curve, normal, xdir, origin)
    else:
        nodes2d = import_polycurve(outer_curve, normal, xdir)

    if nodes2d is None or t is None:
        raise ValueError("Unable to get plate nodes or thickness")

    placement = Placement(origin, xdir=xdir, zdir=normal)

    return Plate(
        name,
        nodes2d,
        t,
        mat=mat,
        placement=placement,
        guid=ifc_elem.GlobalId,
        ifc_store=ifc_store,
        units=ifc_store.assembly.units,
    )
